from typing import Any, List

import webviz_core_components as wcc
from dash import html
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


class Filters(SettingsGroupABC):
    class Ids(StrEnum):
        THP_LABEL = "thp-label"
        WFR_LABEL = "wfr-label"
        GFR_LABEL = "gfr-label"
        ALQ_LABEL = "alq-label"
        THP = "thp"
        WFR = "wfr"
        GFR = "gfr"
        ALQ = "alq"

    def __init__(self) -> None:
        super().__init__("Filters")

    def layout(self) -> List[Any]:
        return [
            html.Label(
                id=self.register_component_unique_id(Filters.Ids.THP_LABEL),
            ),
            wcc.SelectWithLabel(
                id=self.register_component_unique_id(Filters.Ids.THP),
                size=6,
            ),
            html.Label(
                id=self.register_component_unique_id(Filters.Ids.WFR_LABEL),
            ),
            wcc.SelectWithLabel(
                id=self.register_component_unique_id(Filters.Ids.WFR),
                size=6,
            ),
            html.Label(
                id=self.register_component_unique_id(Filters.Ids.GFR_LABEL),
            ),
            wcc.SelectWithLabel(
                id=self.register_component_unique_id(Filters.Ids.GFR),
                size=6,
            ),
            html.Label(
                id=self.register_component_unique_id(Filters.Ids.ALQ_LABEL),
            ),
            wcc.SelectWithLabel(
                id=self.register_component_unique_id(Filters.Ids.ALQ),
                size=6,
            ),
        ]



import warnings
from enum import Enum
from typing import Dict, List

from ._types import LayerTypes


class DeckGLMapLayersModel:
    """Handles updates to the DeckGLMap layers prop"""

    def __init__(self, layers: List[Dict]) -> None:
        self._layers = layers

    def _update_layer_by_type(self, layer_type: Enum, layer_data: Dict) -> None:
        """Update a layer specification by the layer type. If multiple layers are found,
        no update is performed."""
        layers = list(filter(lambda x: x["@@type"] == layer_type, self._layers))
        if not layers:
            warnings.warn(f"No {layer_type} found in layer specification!")
        if len(layers) > 1:
            warnings.warn(
                f"Multiple layers of type {layer_type} found in layer specification!"
            )
        if len(layers) == 1:
            layer_idx = self._layers.index(layers[0])
            self._layers[layer_idx].update(layer_data)

    def update_layer_by_id(self, layer_id: str, layer_data: Dict) -> None:
        """Update a layer specification by the layer id."""
        layers = list(filter(lambda x: x["id"] == layer_id, self._layers))
        if not layers:
            warnings.warn(f"No layer with id {layer_id} found in layer specification!")
        if len(layers) > 1:
            warnings.warn(
                f"Multiple layers with id {layer_id} found in layer specification!"
            )
        if len(layers) == 1:
            layer_idx = self._layers.index(layers[0])
            self._layers[layer_idx].update(layer_data)

    def set_propertymap(
        self,
        image_url: str,
        bounds: List[float],
        value_range: List[float],
    ) -> None:
        """Set the property map image url, bounds and value range in the
        Colormap and Hillshading layer"""
        self._update_layer_by_type(
            layer_type=LayerTypes.HILLSHADING,
            layer_data={
                "image": image_url,
                "bounds": bounds,
                "valueRange": value_range,
            },
        )
        self._update_layer_by_type(
            layer_type=LayerTypes.COLORMAP,
            layer_data={
                "image": image_url,
                "bounds": bounds,
                "valueRange": value_range,
            },
        )

    def set_colormap_image(self, colormap: str) -> None:
        """Set the colormap image url in the ColormapLayer"""
        self._update_layer_by_type(
            layer_type=LayerTypes.COLORMAP,
            layer_data={
                "colormap": colormap,
            },
        )

    def set_colormap_range(self, colormap_range: List[float]) -> None:
        """Set the colormap range in the ColormapLayer"""
        self._update_layer_by_type(
            layer_type=LayerTypes.COLORMAP,
            layer_data={
                "colorMapRange": colormap_range,
            },
        )

    def set_well_data(self, well_data: List[Dict]) -> None:
        """Set the well data json url in the WellsLayer"""
        self._update_layer_by_type(
            layer_type=LayerTypes.WELL,
            layer_data={
                "data": well_data,
            },
        )

    @property
    def layers(self) -> List[Dict]:
        """Returns the full layers specification"""
        return self._layers



import datetime

import pandas as pd
import pytest
from pandas._testing import assert_frame_equal

from webviz_subsurface._utils.dataframe_utils import (
    assert_date_column_is_datetime_object,
    make_date_column_datetime_object,
)

# *******************************************************
#########################################################
#
# SETUP OF INPUT DATAFRAMES FOR TESTS
#
#########################################################
# *******************************************************

# fmt: off
INPUT_EMPTY_DF = pd.DataFrame()
INPUT_NO_ROWS_DF = pd.DataFrame(columns=["DATE", "A"])
INPUT_TIMESTAMP_DATETIME_DF = pd.DataFrame(
    columns=["DATE", "A"],
    data=[
        [pd.Timestamp(2000, 1, 1),       1.0],  # pd.Timestamp detected in df["DATE"][0]
        [datetime.datetime(2263, 2, 15), 1.0],
    ],
)
INPUT_DATETIME_YEAR_2020_DF = pd.DataFrame(
    columns=["DATE", "A"],
    data=[
        [datetime.datetime(2020, 1, 15), 1.0],
        [datetime.datetime(2020, 2, 15), 2.0],
        [datetime.datetime(2020, 3, 15), 3.0],
    ],
)
INPUT_DATETIME_YEAR_2020_DF["DATE"] = pd.Series(
    INPUT_DATETIME_YEAR_2020_DF["DATE"].dt.to_pydatetime(), dtype=object
) # Converting "DATE"-column to datetime.datetime after construct

INPUT_DATETIME_YEAR_2263_DF = pd.DataFrame(
    columns=["DATE", "A"],
    data=[
        [datetime.datetime(2263, 1, 15), 1.0],
        [datetime.datetime(2263, 2, 15), 2.0],
        [datetime.datetime(2263, 3, 15), 3.0],
    ],
)
INPUT_TIMESTAMP_YEAR_2020_DF = pd.DataFrame(
    columns=["DATE", "A"],
    data=[
        [pd.Timestamp(2020, 1, 15), 1.0],
        [pd.Timestamp(2020, 2, 15), 2.0],
        [pd.Timestamp(2020, 3, 15), 3.0],
    ],
)

EXPECTED_YEAR_2020_DF = pd.DataFrame(
    columns=["DATE", "A"],
    data=[
        [datetime.datetime(2020, 1, 15), 1.0],
        [datetime.datetime(2020, 2, 15), 2.0],
        [datetime.datetime(2020, 3, 15), 3.0],
    ],
)
# Convert "DATE" column to datetime.datetime format as DataFrame constructor
# converts datetime.datetime to pd.Timestamp
EXPECTED_YEAR_2020_DF["DATE"] = pd.Series(
    EXPECTED_YEAR_2020_DF["DATE"].dt.to_pydatetime(), dtype=object
)

# Constructor not converting datetime.datetime for year > 2262 to pd.Timestamp
EXPECTED_YEAR_2263_DF = pd.DataFrame(
    columns=["DATE", "A"],
    data=[
        [datetime.datetime(2263, 1, 15), 1.0],
        [datetime.datetime(2263, 2, 15), 2.0],
        [datetime.datetime(2263, 3, 15), 3.0],
    ],
)
# fmt: on

# *******************************************************
#########################################################
#
# TESTING OF: assert_date_column_is_datetime_object()
#
#########################################################
# *******************************************************


def test_assert_date_column_is_datetime_object_no_date_column_error() -> None:
    with pytest.raises(ValueError) as err:
        assert_date_column_is_datetime_object(INPUT_EMPTY_DF)
    assert str(err.value) == 'df does not contain column "DATE"'


def test_assert_date_column_is_datetime_object_timestamp_input_error() -> None:
    # fmt: off
    input_timestamp_df = pd.DataFrame(
        columns=["DATE", "A"],
        data=[
            [pd.Timestamp(2000, 1, 15), 1.0],
            [pd.Timestamp(2000, 2, 15), 1.0]],
    )
    # fmt: on
    with pytest.raises(ValueError) as err:
        assert_date_column_is_datetime_object(input_timestamp_df)
    assert (
        str(err.value)
        == '"DATE"-column in dataframe is not on datetime.datetime format!'
    )


def test_assert_date_column_is_datetime_object_timestamp_datetime_input_error() -> None:
    with pytest.raises(ValueError) as err:
        assert_date_column_is_datetime_object(INPUT_TIMESTAMP_DATETIME_DF)
    assert (
        str(err.value)
        == '"DATE"-column in dataframe is not on datetime.datetime format!'
    )


def test_assert_date_column_is_datetime_object_no_rows_df() -> None:
    try:
        assert_date_column_is_datetime_object(INPUT_NO_ROWS_DF)

    # pylint: disable = bare-except
    except:
        pytest.fail("Excpected no raise of ERROR!")


def test_assert_date_column_is_datetime_object_datetime_year_2020() -> None:
    try:
        assert_date_column_is_datetime_object(INPUT_DATETIME_YEAR_2020_DF)

    # pylint: disable = bare-except
    except:
        pytest.fail("Excpected no raise of ERROR!")


def test_assert_date_column_is_datetime_object_datetime_year_2263() -> None:
    try:
        assert_date_column_is_datetime_object(INPUT_DATETIME_YEAR_2263_DF)

    # pylint: disable = bare-except
    except:
        pytest.fail("Excpected no raise of ERROR!")


def test_assert_date_column_is_datetime_object_datetime_timestamp_df() -> None:
    # fmt: off
    input_datetime_timestamp_df = pd.DataFrame(
        data=[
            [datetime.datetime(2263, 2, 1), 1.0],
            [pd.Timestamp(2000, 1, 1),      1.0],  # pd.Timestamp NOT detected in df["DATE"][1]
        ],
        columns=["DATE", "A"],
    )
    # fmt: on
    try:
        assert_date_column_is_datetime_object(input_datetime_timestamp_df)

    # pylint: disable = bare-except
    except:
        pytest.fail("Excpected no raise of ERROR!")


def test_assert_date_column_is_datetime_object_datetime_inconsistent_index_df() -> None:
    """To verify iloc usage"""
    # fmt: off
    input_datetime_year_2263_inconsistent_index_df = pd.DataFrame(
        columns=["DATE", "A"],
        data=[
            [datetime.datetime(2263, 1, 15), 1.0],
            [datetime.datetime(2263, 2, 15), 2.0],
            [datetime.datetime(2263, 3, 15), 3.0],
        ],
        index = [2,5,9]
    )
    # fmt: on
    try:
        assert_date_column_is_datetime_object(
            input_datetime_year_2263_inconsistent_index_df
        )

    # pylint: disable = bare-except
    except:
        pytest.fail("Excpected no raise of ERROR!")


# *******************************************************
########################################################
#
# TESTING OF: make_date_column_datetime_object()
#
########################################################
# *******************************************************


def _verify_expected_df_date_column_data(df: pd.DataFrame) -> None:
    """Verify dataframe contains column named "DATE" and that
    each row in column is of type datetime.datetime.

    NOTE: Can be 0 rows
    """
    assert "DATE" in df.columns
    for row in df["DATE"]:
        # pylint: disable = unidiomatic-typecheck
        assert type(row) == datetime.datetime


def test_make_date_column_datetime_object_datetime_year_2020_df() -> None:
    _verify_expected_df_date_column_data(EXPECTED_YEAR_2020_DF)

    # Copy to prevent modification if input_df
    test_df = INPUT_DATETIME_YEAR_2020_DF.copy()

    make_date_column_datetime_object(test_df)

    assert_frame_equal(test_df, EXPECTED_YEAR_2020_DF)


def test_make_date_column_datetime_object_timestamp_year_2020_df() -> None:
    _verify_expected_df_date_column_data(EXPECTED_YEAR_2020_DF)

    # Copy to prevent modification if input_df
    test_df = INPUT_TIMESTAMP_YEAR_2020_DF.copy()

    make_date_column_datetime_object(test_df)

    assert_frame_equal(test_df, EXPECTED_YEAR_2020_DF)


def test_make_date_column_datetime_object_datetime_year_2263_df() -> None:
    _verify_expected_df_date_column_data(EXPECTED_YEAR_2263_DF)

    # Copy to prevent modification if input_df
    test_df = INPUT_DATETIME_YEAR_2263_DF.copy()

    make_date_column_datetime_object(test_df)

    assert_frame_equal(test_df, EXPECTED_YEAR_2263_DF)


def test_make_date_column_datetime_object_no_rows_df() -> None:
    # Copy to prevent modification if input_df
    test_df = INPUT_NO_ROWS_DF.copy()

    make_date_column_datetime_object(test_df)

    assert_frame_equal(test_df, INPUT_NO_ROWS_DF)


def test_make_date_column_datetime_object_input_empty_df() -> None:
    with pytest.raises(ValueError) as err:
        make_date_column_datetime_object(INPUT_EMPTY_DF)
    assert str(err.value) == 'df does not contain column "DATE"'


def test_make_date_column_datetime_object_input_timestamp_datetime_df() -> None:
    with pytest.raises(AttributeError) as err:
        make_date_column_datetime_object(INPUT_TIMESTAMP_DATETIME_DF)
    assert str(err.value) == "Can only use .dt accessor with datetimelike values"


def test_make_date_column_datetime_object_input_date_year_2020_df() -> None:
    # fmt: off
    input_date_year_2020_df = pd.DataFrame(
        columns=["DATE", "A"],
        data=[
            [datetime.date(2020, 1, 15), 1.0],
            [datetime.date(2020, 2, 15), 2.0],
            [datetime.date(2020, 3, 15), 3.0],
        ],
    )
    # fmt: on
    with pytest.raises(ValueError) as err:
        make_date_column_datetime_object(input_date_year_2020_df)
    assert str(err.value) == f'Column "DATE" of type {datetime.date} is not handled!'



from pathlib import Path
from typing import Any, Dict, List

import yaml


def load_and_validate_log_templates(log_templates: List[Path]) -> Dict[str, Any]:
    validated_templates = {}
    for idx, template_path in enumerate(log_templates):
        template = yaml.safe_load(template_path.read_text())
        # Validate against json schema here when available
        # https://github.com/equinor/webviz-subsurface-components/issues/508
        template_name = template.get("name", f"template_{idx+1}")
        validated_templates[template_name] = template
    return validated_templates



import hashlib
import json
import logging
import os
from pathlib import Path
from typing import Dict

import pandas as pd
from webviz_config.webviz_factory import WebvizFactory
from webviz_config.webviz_factory_registry import WEBVIZ_FACTORY_REGISTRY
from webviz_config.webviz_instance_info import WebvizRunMode

# The fmu.ensemble dependency ecl is only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    from fmu.ensemble import ScratchEnsemble
except ImportError:
    pass

from webviz_subsurface._utils.perf_timer import PerfTimer

from ..ensemble_summary_provider._arrow_unsmry_import import (
    load_per_realization_arrow_unsmry_files,
)
from ..ensemble_summary_provider._csv_import import load_per_real_csv_file_using_fmu
from .ensemble_table_provider import EnsembleTableProvider
from .ensemble_table_provider_impl_arrow import EnsembleTableProviderImplArrow

LOGGER = logging.getLogger(__name__)


class EnsembleTableProviderFactory(WebvizFactory):
    def __init__(self, root_storage_folder: Path, allow_storage_writes: bool) -> None:
        self._storage_dir = Path(root_storage_folder) / __name__
        self._allow_storage_writes = allow_storage_writes

        LOGGER.info(
            f"EnsembleTableProviderFactory init: storage_dir={self._storage_dir}"
        )

        if self._allow_storage_writes:
            os.makedirs(self._storage_dir, exist_ok=True)

    @staticmethod
    def instance() -> "EnsembleTableProviderFactory":
        """Static method to access the singleton instance of the factory."""

        factory = WEBVIZ_FACTORY_REGISTRY.get_factory(EnsembleTableProviderFactory)
        if not factory:
            app_instance_info = WEBVIZ_FACTORY_REGISTRY.app_instance_info
            storage_folder = app_instance_info.storage_folder
            allow_writes = app_instance_info.run_mode != WebvizRunMode.PORTABLE

            factory = EnsembleTableProviderFactory(storage_folder, allow_writes)

            # Store the factory object in the global factory registry
            WEBVIZ_FACTORY_REGISTRY.set_factory(EnsembleTableProviderFactory, factory)

        return factory

    def create_from_ensemble_csv_file(
        self,
        csv_file: Path,
    ) -> EnsembleTableProvider:
        """Create EnsembleTableProvider from aggregated CSV file.
        The CSV file is assumed to contain data for a single ensemble and must contain
        a REAL column.
        If the CSV file contains an `ENSEMBLE` column it will be ignored, but an exception
        will be thrown if it is present and it contains multiple ensemble names.
        """

        timer = PerfTimer()

        storage_key = f"ens_csv__{_make_hash_string(str(csv_file))}"

        provider = EnsembleTableProviderImplArrow.from_backing_store(
            self._storage_dir, storage_key
        )
        if provider:
            LOGGER.info(
                f"Loaded table provider (CSV) from backing store in "
                f"{timer.elapsed_s():.2f}s (csv_file={csv_file})"
            )
            return provider

        # We can only import data from csv if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(f"Failed to load table provider (CSV) for {csv_file}")

        LOGGER.info(f"Importing/saving CSV data for: {csv_file}")

        timer.lap_s()
        ensemble_df = pd.read_csv(csv_file)

        if "ENSEMBLE" in ensemble_df.columns:
            if ensemble_df["ENSEMBLE"].nunique() > 1:
                raise KeyError("Input data contains more than one unique ensemble name")

        et_import_csv_s = timer.lap_s()

        if ensemble_df.empty:
            raise ValueError("Import resulted in empty DataFrame")
        if "REAL" not in ensemble_df.columns:
            raise ValueError("No REAL column present in input data")

        EnsembleTableProviderImplArrow.write_backing_store_from_ensemble_dataframe(
            self._storage_dir, storage_key, ensemble_df
        )
        et_write_s = timer.lap_s()

        provider = EnsembleTableProviderImplArrow.from_backing_store(
            self._storage_dir, storage_key
        )
        if not provider:
            raise ValueError(f"Failed to load/create provider for {csv_file}")

        LOGGER.info(
            f"Saved table provider (CSV) to backing store in {timer.elapsed_s():.2f}s ("
            f"import_csv={et_import_csv_s:.2f}s, "
            f"write={et_write_s:.2f}s, "
            f"csv_file={csv_file})"
        )

        return provider

    def create_from_per_realization_csv_file(
        self, ens_path: str, csv_file_rel_path: str
    ) -> EnsembleTableProvider:
        """Create EnsembleTableProvider from per realization CSV files.

        Note that the returned table provider will not be able to return vector
        metadata.
        """

        timer = PerfTimer()

        storage_key = f"per_real_csv__{_make_hash_string(ens_path + csv_file_rel_path)}"
        provider = EnsembleTableProviderImplArrow.from_backing_store(
            self._storage_dir, storage_key
        )

        if provider:
            LOGGER.info(
                f"Loaded table provider (per real CSV) from backing store in "
                f"{timer.elapsed_s():.2f}s ("
                f"ens_path={ens_path}, csv_file_rel_path={csv_file_rel_path})"
            )
            return provider

        # We can only import data from CSV if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(
                f"Failed to load table provider (per real CSV) for {ens_path}"
            )

        LOGGER.info(f"Importing/saving per real CSV data for: {ens_path}")

        timer.lap_s()

        ensemble_df = load_per_real_csv_file_using_fmu(ens_path, csv_file_rel_path)
        et_import_csv_s = timer.lap_s()

        EnsembleTableProviderImplArrow.write_backing_store_from_ensemble_dataframe(
            self._storage_dir, storage_key, ensemble_df
        )
        et_write_s = timer.lap_s()

        provider = EnsembleTableProviderImplArrow.from_backing_store(
            self._storage_dir, storage_key
        )

        if not provider:
            raise ValueError(
                f"Failed to load/create provider (per real CSV) for {ens_path}"
            )

        LOGGER.info(
            f"Saved table provider (per real CSV) to backing store in {timer.elapsed_s():.2f}s ("
            f"import_csv={et_import_csv_s:.2f}s, write={et_write_s:.2f}s, "
            f"ens_path={ens_path}, csv_file_rel_path={csv_file_rel_path})"
        )

        return provider

    def create_from_per_realization_arrow_file(
        self, ens_path: str, rel_file_pattern: str
    ) -> EnsembleTableProvider:
        """Create EnsembleTableProvider from per realization data in .arrow format.

        The `rel_file_pattern` parameter must specify a relative (per realization) file pattern
        that will be used to find the wanted .arrow files within each realization. The file
        pattern is realtive to each realizations's `runpath`.
        """

        timer = PerfTimer()

        storage_key = (
            f"per_real_arrow__{_make_hash_string(ens_path + rel_file_pattern)}"
        )
        provider = EnsembleTableProviderImplArrow.from_backing_store(
            self._storage_dir, storage_key
        )
        if provider:
            LOGGER.info(
                f"Loaded table provider from backing store in {timer.elapsed_s():.2f}s ("
                f"ens_path={ens_path})"
            )
            return provider

        # We can only import data from data source if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(f"Failed to load table provider for {ens_path}")

        LOGGER.info(f"Importing/saving arrow table data for: {ens_path}")

        timer.lap_s()
        per_real_tables = load_per_realization_arrow_unsmry_files(
            ens_path, rel_file_pattern
        )
        if not per_real_tables:
            raise ValueError(f"Could not find any .arrow files for ens_path={ens_path}")
        et_import_smry_s = timer.lap_s()

        try:
            EnsembleTableProviderImplArrow.write_backing_store_from_per_realization_tables(
                self._storage_dir, storage_key, per_real_tables
            )
        except ValueError as exc:
            raise ValueError(f"Failed to write backing store for: {ens_path}") from exc

        et_write_s = timer.lap_s()

        provider = EnsembleTableProviderImplArrow.from_backing_store(
            self._storage_dir, storage_key
        )
        if not provider:
            raise ValueError(f"Failed to load/create table provider for {ens_path}")

        LOGGER.info(
            f"Saved table provider to backing store in {timer.elapsed_s():.2f}s ("
            f"import_smry={et_import_smry_s:.2f}s, write={et_write_s:.2f}s, ens_path={ens_path})"
        )

        return provider

    def create_from_per_realization_parameter_file(
        self, ens_path: str
    ) -> EnsembleTableProvider:
        """Create EnsembleTableProvider from parameter files.

        Note that the returned table provider will not be able to return metadata.
        """

        LOGGER.info("create_provider_from_per_realization_parameter_file() ...")

        timer = PerfTimer()

        storage_key = f"parameters_{_make_hash_string(ens_path + '_parameters')}"

        provider = EnsembleTableProviderImplArrow.from_backing_store(
            self._storage_dir, storage_key
        )

        if provider:
            LOGGER.info(
                f"Loaded table provider from backing store in {timer.elapsed_s():.2f}s ("
                f"ens_path={ens_path})"
            )
            return provider

        # We can only import data from data source if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(f"Failed to load table provider for {ens_path}")

        LOGGER.info(f"Importing parameters for: {ens_path}")

        timer.lap_s()

        scratch_ensemble = ScratchEnsemble("ens_name_tmp", ens_path).filter("OK")
        ensemble_df = scratch_ensemble.parameters
        del scratch_ensemble
        elapsed_load_parameters_s = timer.lap_s()

        try:
            EnsembleTableProviderImplArrow.write_backing_store_from_ensemble_dataframe(
                self._storage_dir, storage_key, ensemble_df
            )
        except ValueError as exc:
            raise ValueError(f"Failed to write backing store for: {ens_path}") from exc

        et_write_s = timer.lap_s()

        provider = EnsembleTableProviderImplArrow.from_backing_store(
            self._storage_dir, storage_key
        )
        if not provider:
            raise ValueError(f"Failed to load/crate table provider for {ens_path}")

        LOGGER.info(
            f"Saved table provider to backing store in {timer.elapsed_s():.2f}s ("
            f"load_parameters={elapsed_load_parameters_s:.2f}s, "
            f"write={et_write_s:.2f}s, ens_path={ens_path})"
        )

        return provider

    def create_provider_set_from_aggregated_csv_file(
        self,
        aggr_csv_file: Path,
    ) -> Dict[str, EnsembleTableProvider]:
        """Creates a dictionary of table providers from an aggregated CSV file with
        ENSEMBLE column. The CSV file can have multiple ensembles in the ENSEMBLE column.
        (This is not accepted by the create_from_ensemble_csv_file function)

        Aggregated csv-files per ensemble is the preferred method.
        """
        LOGGER.info(f"create_provider_set_from_aggregated_csv_file() - {aggr_csv_file}")

        hashval = _make_hash_string(str(aggr_csv_file))
        main_storage_key = f"aggr_csv__{hashval}"

        storage_keys_to_load: Dict[str, str] = {}
        json_fn = self._storage_dir / (f"{main_storage_key}.json")
        try:
            with open(json_fn, "r") as file:
                storage_keys_to_load = json.load(file)
        except FileNotFoundError:
            # We can only recover from this if we're allowed to write to storage
            if not self._allow_storage_writes:
                raise

        if not storage_keys_to_load and self._allow_storage_writes:
            aggregated_df = pd.read_csv(aggr_csv_file)
            ensemble_names = aggregated_df["ENSEMBLE"].unique()

            LOGGER.info(
                f"Saving {len(ensemble_names)} table providers from aggregated CSV to backing store"
            )

            for ens_name in ensemble_names:
                storage_key = f"{main_storage_key}__{ens_name}"
                ensemble_df = aggregated_df[aggregated_df["ENSEMBLE"] == ens_name]
                EnsembleTableProviderImplArrow.write_backing_store_from_ensemble_dataframe(
                    self._storage_dir, storage_key, ensemble_df
                )
                storage_keys_to_load[ens_name] = storage_key

            with open(json_fn, "w") as file:
                json.dump(storage_keys_to_load, file)

        created_providers: Dict[str, EnsembleTableProvider] = {}
        for ens_name, storage_key in storage_keys_to_load.items():
            provider = EnsembleTableProviderImplArrow.from_backing_store(
                self._storage_dir, storage_key
            )
            if provider:
                created_providers[ens_name] = provider

        num_missing_models = len(storage_keys_to_load) - len(created_providers)
        if num_missing_models > 0:
            raise ValueError(f"Failed to load data for {num_missing_models} ensembles")

        LOGGER.info(f"Loaded {len(created_providers)} providers from backing store")

        return created_providers


def _make_hash_string(string_to_hash: str) -> str:
    # There is no security risk here and chances of collision should be very slim
    return hashlib.md5(string_to_hash.encode()).hexdigest()  # nosec



from typing import Dict, List, Union

import webviz_core_components as wcc
from dash import Input, Output, callback
from dash.development.base_component import Component
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC, ViewABC

from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)

from .._plugin_ids import PluginIds
from ..utils import make_dataframes as makedf
from ..utils import make_figures as makefigs
from ._view_functions import _get_well_names_combined


class PlotSettingsCoverage(SettingsGroupABC):
    class Ids:
        # pylint: disable=too-few-public-methods
        COLORBY = "colorby"
        FIG_LAYOUT_HEIGHT = "fig-layout-height"
        PLOT_TYPE = "plot-type"
        COLORBY_GROUPING = "colorby-grouping"
        SHOW_POINTS = "show-points"

    def __init__(self) -> None:
        super().__init__("Plot Settings")

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Color by",
                id=self.register_component_unique_id(PlotSettingsCoverage.Ids.COLORBY),
                options=[
                    {
                        "label": "Ensemble",
                        "value": "ENSEMBLE",
                    },
                    # {"label": "Well", "value": "WELL"},
                    {"label": "Date", "value": "DATE"},
                ],
                value="ENSEMBLE",
                multi=False,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
            wcc.Dropdown(
                label="Plot type",
                id=self.register_component_unique_id(
                    PlotSettingsCoverage.Ids.PLOT_TYPE
                ),
                options=[
                    {"label": "Diff plot", "value": "diffplot"},
                    {
                        "label": "Diff plot relative (%)",
                        "value": "rel_diffplot",
                    },
                    {"label": "Cross plot", "value": "crossplot"},
                ],
                value="diffplot",
                multi=False,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
            wcc.Dropdown(
                label="Fig layout - height",
                id=self.register_component_unique_id(
                    PlotSettingsCoverage.Ids.FIG_LAYOUT_HEIGHT
                ),
                options=[
                    {
                        "label": "Very small",
                        "value": 250,
                    },
                    {
                        "label": "Small",
                        "value": 350,
                    },
                    {
                        "label": "Medium",
                        "value": 450,
                    },
                    {
                        "label": "Large",
                        "value": 700,
                    },
                    {
                        "label": "Very large",
                        "value": 1000,
                    },
                ],
                value=450,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
            wcc.RadioItems(
                label="Grouping",
                id=self.register_component_unique_id(
                    PlotSettingsCoverage.Ids.COLORBY_GROUPING
                ),
                options=[
                    {"label": "Side by side", "value": "group"},
                    {"label": "Overlay", "value": "overlay"},
                ],
                value="group",
            ),
            wcc.RadioItems(
                label="Show points",
                id=self.register_component_unique_id(
                    PlotSettingsCoverage.Ids.SHOW_POINTS
                ),
                options=[
                    {"label": "Outliers only", "value": "outliers"},
                    {"label": "All points", "value": "all"},
                    {
                        "label": "All points, no box",
                        "value": "strip",
                    },
                ],
                value="outliers",
            ),
        ]


class ProdCoverageView(ViewABC):
    class Ids:
        # pylint: disable=too-few-public-methods
        PLOT_SETTINGS = "plot-settings"
        MAIN_COLUMN = "main-column"

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_provider_set: EnsembleSummaryProviderSet,
        ens_vectors: Dict[str, List[str]],
        ens_realizations: Dict[str, List[int]],
        well_collections: Dict[str, List[str]],
    ) -> None:
        super().__init__("Well production coverage")

        self.input_provider_set = input_provider_set
        self.ens_vectors = ens_vectors
        self.ens_realizations = ens_realizations
        self.well_collections = well_collections

        self.add_settings_group(
            PlotSettingsCoverage(), ProdCoverageView.Ids.PLOT_SETTINGS
        )
        self.main_column = self.add_column(ProdCoverageView.Ids.MAIN_COLUMN)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.layout_element(ProdCoverageView.Ids.MAIN_COLUMN)
                .get_unique_id()
                .to_string(),
                "children",
            ),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_ENSEMBLES), "data"
            ),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_DATES), "data"),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_PHASE), "data"),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_WELLS), "data"),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_WELL_COLLECTIONS),
                "data",
            ),
            Input(
                self.get_store_unique_id(
                    PluginIds.Stores.SELECTED_COMBINE_WELLS_COLLECTION
                ),
                "data",
            ),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_REALIZATIONS), "data"
            ),
            Input(
                self.settings_group(ProdCoverageView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsCoverage.Ids.COLORBY)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(ProdCoverageView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsCoverage.Ids.PLOT_TYPE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(ProdCoverageView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsCoverage.Ids.FIG_LAYOUT_HEIGHT)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(ProdCoverageView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsCoverage.Ids.COLORBY_GROUPING)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(ProdCoverageView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsCoverage.Ids.SHOW_POINTS)
                .to_string(),
                "value",
            ),
            # prevent_initial_call=True,
        )
        def _update_plots(
            ensemble_names: List[str],
            selector_dates: List[str],
            selector_phases: List[str],
            selector_well_names: List[str],
            selector_well_collection_names: List[str],
            selector_well_combine_type: str,
            selector_realizations: List[int],
            colorby: str,
            plot_type: str,
            figheight: int,
            boxmode: str,
            boxplot_points: str,
        ) -> Union[str, List[Component]]:

            if not ensemble_names:
                return "No ensembles selected"

            well_names = _get_well_names_combined(
                self.well_collections,
                selector_well_collection_names,
                selector_well_names,
                selector_well_combine_type,
            )

            if plot_type in ["diffplot", "rel_diffplot"]:
                relative_diff = plot_type == "rel_diffplot"
                dframe = makedf.get_df_diff(
                    makedf.get_df_smry(
                        self.input_provider_set,
                        ensemble_names,
                        self.ens_vectors,
                        self.ens_realizations,
                        selector_realizations,
                        well_names,
                        selector_phases,
                        selector_dates,
                    ),
                    relative_diff=relative_diff,
                )
                figures = makefigs.coverage_diffplot(
                    dframe,
                    selector_phases,
                    colorby,
                    vector_type="well",
                    figheight=figheight,
                    boxmode=boxmode,
                    boxplot_points=boxplot_points,
                )
            if plot_type == "crossplot":
                dframe = makedf.get_df_smry(
                    self.input_provider_set,
                    ensemble_names,
                    self.ens_vectors,
                    self.ens_realizations,
                    selector_realizations,
                    well_names,
                    selector_phases,
                    selector_dates,
                )
                figures = makefigs.coverage_crossplot(
                    dframe,
                    selector_phases,
                    colorby,
                    vector_type="well",
                    figheight=figheight,
                    boxplot_points=boxplot_points,
                )

            return figures



from ._types import QualifiedDiffSurfaceAddress, QualifiedSurfaceAddress
from .ensemble_surface_provider import (
    EnsembleSurfaceProvider,
    ObservedSurfaceAddress,
    SimulatedSurfaceAddress,
    StatisticalSurfaceAddress,
    SurfaceAddress,
)
from .ensemble_surface_provider_factory import EnsembleSurfaceProviderFactory
from .surface_array_server import SurfaceArrayMeta, SurfaceArrayServer
from .surface_image_server import SurfaceImageMeta, SurfaceImageServer



import hashlib
import io
import json
import logging
import math
from dataclasses import asdict, dataclass
from typing import List, Optional, Tuple, Union
from urllib.parse import quote
from uuid import uuid4

import flask
import flask_caching
import xtgeo
from dash import Dash
from webviz_config.webviz_instance_info import WEBVIZ_INSTANCE_INFO

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._surface_to_float32_array import surface_to_float32_array
from ._types import QualifiedDiffSurfaceAddress, QualifiedSurfaceAddress
from .ensemble_surface_provider import (
    ObservedSurfaceAddress,
    SimulatedSurfaceAddress,
    StatisticalSurfaceAddress,
    SurfaceAddress,
)

LOGGER = logging.getLogger(__name__)

_ROOT_URL_PATH = "/SurfaceArrayServer"

_SURFACE_SERVER_INSTANCE: Optional["SurfaceArrayServer"] = None


@dataclass(frozen=True)
class SurfaceArrayMeta:
    x_min: float
    x_max: float
    y_min: float
    y_max: float
    x_ori: float
    y_ori: float
    val_min: float
    val_max: float
    rot_deg: float
    x_count: int
    y_count: int
    x_inc: float
    y_inc: float


class SurfaceArrayServer:
    def __init__(self, app: Dash) -> None:
        cache_dir = (
            WEBVIZ_INSTANCE_INFO.storage_folder
            / f"SurfaceArrayServer_filecache_{uuid4()}"
        )
        LOGGER.debug(f"Setting up file cache in: {cache_dir}")
        self._array_cache = flask_caching.Cache(
            config={
                "CACHE_TYPE": "FileSystemCache",
                "CACHE_DIR": cache_dir,
                "CACHE_DEFAULT_TIMEOUT": 0,
            }
        )
        self._array_cache.init_app(app.server)

        self._setup_url_rule(app)

    @staticmethod
    def instance(app: Dash) -> "SurfaceArrayServer":
        # pylint: disable=global-statement
        global _SURFACE_SERVER_INSTANCE
        if not _SURFACE_SERVER_INSTANCE:
            LOGGER.debug("Initializing SurfaceArrayServer instance")
            _SURFACE_SERVER_INSTANCE = SurfaceArrayServer(app)

        return _SURFACE_SERVER_INSTANCE

    def publish_surface(
        self,
        qualified_address: Union[QualifiedSurfaceAddress, QualifiedDiffSurfaceAddress],
        surface: xtgeo.RegularSurface,
    ) -> None:
        timer = PerfTimer()

        if isinstance(qualified_address, QualifiedSurfaceAddress):
            base_cache_key = _address_to_str(
                qualified_address.provider_id, qualified_address.address
            )
        else:
            base_cache_key = _diff_address_to_str(
                qualified_address.provider_id_a,
                qualified_address.address_a,
                qualified_address.provider_id_b,
                qualified_address.address_b,
            )

        LOGGER.debug(
            f"Publishing surface (dim={surface.dimensions}, #cells={surface.ncol*surface.nrow}), "
            f"[base_cache_key={base_cache_key}]"
        )

        self._create_and_store_array_in_cache(base_cache_key, surface)

        LOGGER.debug(f"Surface published in: {timer.elapsed_s():.2f}s")

    def get_surface_metadata(
        self,
        qualified_address: Union[QualifiedSurfaceAddress, QualifiedDiffSurfaceAddress],
    ) -> Optional[SurfaceArrayMeta]:

        if isinstance(qualified_address, QualifiedSurfaceAddress):
            base_cache_key = _address_to_str(
                qualified_address.provider_id, qualified_address.address
            )
        else:
            base_cache_key = _diff_address_to_str(
                qualified_address.provider_id_a,
                qualified_address.address_a,
                qualified_address.provider_id_b,
                qualified_address.address_b,
            )

        meta_cache_key = "META:" + base_cache_key
        meta: Optional[SurfaceArrayMeta] = self._array_cache.get(meta_cache_key)
        if not meta:
            return None

        if not isinstance(meta, SurfaceArrayMeta):
            LOGGER.error("Error loading SurfaceArrayMeta from cache")
            return None

        return meta

    @staticmethod
    def encode_partial_url(
        qualified_address: Union[QualifiedSurfaceAddress, QualifiedDiffSurfaceAddress],
    ) -> str:

        if isinstance(qualified_address, QualifiedSurfaceAddress):
            address_str = _address_to_str(
                qualified_address.provider_id, qualified_address.address
            )
        else:
            address_str = _diff_address_to_str(
                qualified_address.provider_id_a,
                qualified_address.address_a,
                qualified_address.provider_id_b,
                qualified_address.address_b,
            )

        url_path: str = f"{_ROOT_URL_PATH}/{quote(address_str)}"
        return url_path

    def _setup_url_rule(self, app: Dash) -> None:
        @app.server.route(_ROOT_URL_PATH + "/<full_surf_address_str>")
        def _handle_surface_request(full_surf_address_str: str) -> flask.Response:
            LOGGER.debug(
                f"Handling surface_request: "
                f"full_surf_address_str={full_surf_address_str} "
            )

            timer = PerfTimer()

            array_cache_key = "ARRAY:" + full_surf_address_str
            LOGGER.debug(f"Looking for array in cache (key={array_cache_key}")

            cached_array_bytes = self._array_cache.get(array_cache_key)
            if not cached_array_bytes:
                LOGGER.error(
                    f"Error getting array for address: {full_surf_address_str}"
                )
                flask.abort(404)

            response = flask.send_file(
                cached_array_bytes, mimetype="application/octet-stream"
            )
            LOGGER.debug(
                f"Request handled from array cache in: {timer.elapsed_s():.2f}s"
            )
            return response

    def _create_and_store_array_in_cache(
        self,
        base_cache_key: str,
        surface: xtgeo.RegularSurface,
    ) -> None:

        timer = PerfTimer()
        LOGGER.debug("Converting surface to float32 array...")
        array_bytes: io.BytesIO = surface_to_float32_array(surface)

        et_to_array_s = timer.lap_s()

        array_cache_key = "ARRAY:" + base_cache_key
        meta_cache_key = "META:" + base_cache_key

        self._array_cache.add(array_cache_key, array_bytes)

        meta = SurfaceArrayMeta(
            x_min=surface.xmin,
            x_max=surface.xmax,
            y_min=surface.ymin,
            y_max=surface.ymax,
            x_ori=surface.xori,
            y_ori=surface.yori,
            x_count=surface.ncol,
            y_count=surface.nrow,
            val_min=surface.values.min(),
            val_max=surface.values.max(),
            rot_deg=surface.rotation,
            x_inc=surface.xinc,
            y_inc=surface.yinc,
        )
        self._array_cache.add(meta_cache_key, meta)
        et_write_cache_s = timer.lap_s()

        LOGGER.debug(
            f"Created surface array and wrote to cache in in: {timer.elapsed_s():.2f}s ("
            f"to_array={et_to_array_s:.2f}s, write_cache={et_write_cache_s:.2f}s), "
            f"[base_cache_key={base_cache_key}]"
        )


def _address_to_str(
    provider_id: str,
    address: SurfaceAddress,
) -> str:
    if isinstance(address, StatisticalSurfaceAddress):
        addr_type_str = "sta"
    elif isinstance(address, SimulatedSurfaceAddress):
        addr_type_str = "sim"
    elif isinstance(address, ObservedSurfaceAddress):
        addr_type_str = "obs"

    addr_hash = hashlib.md5(  # nosec
        json.dumps(asdict(address), sort_keys=True).encode()
    ).hexdigest()

    return f"{provider_id}___{addr_type_str}___{address.name}___{address.attribute}___{addr_hash}"


def _diff_address_to_str(
    provider_id_a: str,
    address_a: SurfaceAddress,
    provider_id_b: str,
    address_b: SurfaceAddress,
) -> str:
    return (
        "diff~~~"
        + _address_to_str(provider_id_a, address_a)
        + "~~~"
        + _address_to_str(provider_id_b, address_b)
    )


def _calc_map_component_bounds_and_rot(
    surface: xtgeo.RegularSurface,
) -> Tuple[List[float], float]:
    surf_corners = surface.get_map_xycorners()
    rptx = surf_corners[2][0]
    rpty = surf_corners[2][1]
    min_x = math.inf
    max_x = -math.inf
    min_y = math.inf
    max_y = -math.inf
    angle = -surface.rotation * math.pi / 180
    for coord in surf_corners:
        xpos = coord[0]
        ypos = coord[1]
        x_rotated = (
            rptx + ((xpos - rptx) * math.cos(angle)) - ((ypos - rpty) * math.sin(angle))
        )
        y_rotated = (
            rpty + ((xpos - rptx) * math.sin(angle)) + ((ypos - rpty) * math.cos(angle))
        )
        min_x = min(min_x, x_rotated)
        max_x = max(max_x, x_rotated)
        min_y = min(min_y, y_rotated)
        max_y = max(max_y, y_rotated)

    bounds = [
        min_x,
        min_y,
        max_x,
        max_y,
    ]

    return bounds, surface.rotation



import jwt
import orjson as json
from dash import Dash
from jwt.exceptions import DecodeError

from webviz_subsurface._providers.ensemble_grid_provider import (
    CellFilter,
    GridVizService,
    PropertySpec,
)


def set_routes(app: Dash, grid_viz_service: GridVizService) -> None:
    @app.server.route("/grid/points/<token>")
    def _get_points(token: str) -> bytes:
        geometry_token = jwt.decode(token, "secret", algorithms=["HS256"])
        try:
            geometry_token = jwt.decode(token, "secret", algorithms=["HS256"])
        except DecodeError:
            # # pylint: disable=no-member
            return json.dumps([])
        provider_id = geometry_token["provider_id"]
        realization = geometry_token["realization"]
        # property_spec = PropertySpec(**geometry_token["property_spec"])
        cell_filter = CellFilter(**geometry_token["cell_filter"])
        surface_polys, _ = grid_viz_service.get_surface(
            provider_id=provider_id,
            realization=realization,
            cell_filter=cell_filter,
            property_spec=None,
        )
        # pylint: disable=no-member
        return json.dumps(surface_polys.point_arr.tolist())

    @app.server.route("/grid/polys/<token>")
    def _get_polys(token: str) -> bytes:
        try:
            geometry_token = jwt.decode(token, "secret", algorithms=["HS256"])
        except DecodeError:
            # pylint: disable=no-member
            return json.dumps([])
        provider_id = geometry_token["provider_id"]
        realization = geometry_token["realization"]
        # property_spec = PropertySpec(**geometry_token["property_spec"])
        cell_filter = CellFilter(**geometry_token["cell_filter"])
        surface_polys, _ = grid_viz_service.get_surface(
            provider_id=provider_id,
            realization=realization,
            cell_filter=cell_filter,
            property_spec=None,
        )
        # pylint: disable=no-member
        return json.dumps(surface_polys.poly_arr.tolist())

    @app.server.route("/grid/scalar/<token>")
    def _get_scalar(token: str) -> bytes:
        try:
            geometry_token = jwt.decode(token, "secret", algorithms=["HS256"])
        except DecodeError:
            # pylint: disable=no-member
            return json.dumps([])
        provider_id = geometry_token["provider_id"]
        realization = geometry_token["realization"]
        property_spec = PropertySpec(**geometry_token["property_spec"])
        cell_filter = CellFilter(**geometry_token["cell_filter"])
        scalars = grid_viz_service.get_mapped_property_values(
            provider_id=provider_id,
            realization=realization,
            cell_filter=cell_filter,
            property_spec=property_spec,
        )
        # pylint: disable=no-member
        return json.dumps(scalars.value_arr.tolist() if scalars else [])



from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd


class TornadoData:
    REQUIRED_COLUMNS = ["REAL", "SENSNAME", "SENSCASE", "SENSTYPE", "VALUE"]

    def __init__(
        self,
        dframe: pd.DataFrame,
        response_name: Optional[str] = "Response",
        reference: str = "rms_seed",
        cutbyref: bool = False,
        scale: str = "Percentage",
    ) -> None:
        self._reference = reference
        self.response_name = response_name
        self._validate_input(dframe)
        self._scale = scale
        self._reference_average = self._calculate_ref_average(dframe)
        self._tornadotable = self._calculate_tornado_table(dframe)
        if cutbyref:
            self._cut_sensitivities_by_ref()
        self._sort_sensitivities_by_max()
        self._real_df = self._create_real_df(dframe)

    def _validate_input(self, dframe: pd.DataFrame) -> None:
        for col in self.REQUIRED_COLUMNS:
            if col not in dframe:
                raise KeyError(f"Tornado input is missing {col}")

        if list(dframe["SENSCASE"].unique()) == [None]:
            raise KeyError("No sensitivities found in tornado input")

        for sens_name, sens_name_df in dframe.groupby("SENSNAME"):
            if not any(
                (sens_name_df["SENSTYPE"] == st).all() for st in ["scalar", "mc"]
            ):
                raise ValueError(
                    f"Sensitivity {sens_name} is not of type 'mc' or 'scalar"
                )
        if dframe.loc[dframe["SENSNAME"].isin([self._reference])].empty:
            raise ValueError(f"Reference SENSNAME {self._reference} not in input data")

    def _create_real_df(self, dframe: pd.DataFrame) -> pd.DataFrame:
        """Make dataframe with value and case info per realization"""
        realdf = dframe[self.REQUIRED_COLUMNS].rename(
            columns={"SENSNAME": "sensname", "SENSCASE": "senscase"}
        )

        sensitivities = self._tornadotable["sensname"].unique()
        realdf = realdf.loc[realdf["sensname"].isin(sensitivities)]

        for val in self.low_high_realizations_list.values():
            for case in ["high", "low"]:
                casemask = realdf["REAL"].isin(val[f"real_{case}"])
                realdf.loc[casemask, "case"] = case

        mc_mask = realdf["SENSTYPE"] == "mc"
        realdf["casetype"] = np.where(mc_mask, "mc", realdf["case"])
        realdf["sensname_case"] = np.where(
            mc_mask,
            realdf["sensname"],
            realdf[["sensname", "senscase"]].agg("--".join, axis=1),
        )
        return realdf

    @property
    def real_df(self) -> pd.DataFrame:
        return self._real_df

    @property
    def scale(self) -> str:
        return self._scale

    def _calculate_ref_average(self, dframe: pd.DataFrame) -> float:
        # Calculate average response value for reference sensitivity
        return dframe.loc[dframe["SENSNAME"] == self._reference]["VALUE"].mean()

    @property
    def reference_average(self) -> float:
        return self._reference_average

    def _calculate_tornado_table(self, dframe: pd.DataFrame) -> pd.DataFrame:
        avg_per_sensitivity = self._calculate_sensitivity_averages(dframe)
        return pd.DataFrame(self._calculate_tornado_low_high_list(avg_per_sensitivity))

    @property
    def tornadotable(self) -> pd.DataFrame:
        return self._tornadotable

    def _scale_to_ref(self, value: float) -> float:
        value_ref = value - self.reference_average
        if self.scale == "Percentage":
            return (
                (100 * (value_ref / self.reference_average))
                if self.reference_average != 0
                else 0
            )
        return value_ref

    def _calculate_sensitivity_averages(
        self, dframe: pd.DataFrame
    ) -> List[Dict[str, Union[str, list, float]]]:
        avg_per_sensitivity = []

        for sens_name, sens_name_df in dframe.groupby("SENSNAME"):
            # Excluding cases if `ref` is used as `SENSNAME`, and only one realization
            # is present for this `SENSNAME`
            if sens_name == "ref" and len(sens_name_df["REAL"].unique()) == 1:
                continue

            # If `SENSTYPE` is scalar get the mean for each `SENSCASE`
            if (sens_name_df["SENSTYPE"] == "scalar").all():
                for sens_case, sens_case_df in sens_name_df.groupby("SENSCASE"):
                    avg_per_sensitivity.append(
                        {
                            "sensname": sens_name,
                            "senscase": sens_case,
                            "values": sens_case_df["VALUE"].mean(),
                            "values_ref": self._scale_to_ref(
                                sens_case_df["VALUE"].mean()
                            ),
                            "reals": list(map(int, sens_case_df["REAL"])),
                        }
                    )
            # If `SENSTYPE` is monte carlo get p10, p90
            elif (sens_name_df["SENSTYPE"] == "mc").all():

                # Calculate p90(low) and p10(high)
                p90 = sens_name_df["VALUE"].quantile(0.10)
                p10 = sens_name_df["VALUE"].quantile(0.90)

                # Extract list of realizations with values less then reference avg (low)
                low_reals = list(
                    map(
                        int,
                        sens_name_df.loc[
                            sens_name_df["VALUE"] <= self.reference_average
                        ]["REAL"],
                    )
                )

                # Extract list of realizations with values higher then reference avg (high)
                high_reals = list(
                    map(
                        int,
                        sens_name_df.loc[
                            sens_name_df["VALUE"] > self.reference_average
                        ]["REAL"],
                    )
                )

                avg_per_sensitivity.append(
                    {
                        "sensname": sens_name,
                        "senscase": "P90",
                        "values": p90,
                        "values_ref": self._scale_to_ref(p90),
                        "reals": low_reals,
                    }
                )
                avg_per_sensitivity.append(
                    {
                        "sensname": sens_name,
                        "senscase": "P10",
                        "values": p10,
                        "values_ref": self._scale_to_ref(p10),
                        "reals": high_reals,
                    }
                )

        return avg_per_sensitivity

    def _calculate_tornado_low_high_list(
        self, avg_per_sensitivity: List
    ) -> List[Dict[str, Union[str, list, float]]]:
        low_high_per_sensitivity = []
        for sensname, sens_name_df in pd.DataFrame(avg_per_sensitivity).groupby(
            "sensname"
        ):
            low = sens_name_df.copy().loc[sens_name_df["values_ref"].idxmin()]
            high = sens_name_df.copy().loc[sens_name_df["values_ref"].idxmax()]
            if sens_name_df["senscase"].nunique() == 1:
                # Single case sens, implies low == high, but testing just in case:
                if low["values_ref"] != high["values_ref"]:
                    raise ValueError(
                        "For a single sensitivity case, low and high cases should be equal."
                    )
                if low["values_ref"] < 0:
                    # To avoid warnings for changing values of dataframe slices.
                    high = high.copy()
                    high["values_ref"] = 0
                    high["reals"] = []
                    high["senscase"] = None
                    high["values"] = self.reference_average
                else:
                    low = (
                        low.copy()
                    )  # To avoid warnings for changing values of dataframe slices.
                    low["values_ref"] = 0
                    low["reals"] = []
                    low["senscase"] = None
                    low["values"] = self.reference_average

            low_high_per_sensitivity.append(
                {
                    "low": self.calc_low_x(low["values_ref"], high["values_ref"]),
                    "low_base": self.calc_low_base(
                        low["values_ref"], high["values_ref"]
                    ),
                    "low_label": low["senscase"],
                    "low_tooltip": low["values_ref"],
                    "true_low": low["values"],
                    "low_reals": low["reals"],
                    "sensname": sensname,
                    "high": self.calc_high_x(low["values_ref"], high["values_ref"]),
                    "high_base": self.calc_high_base(
                        low["values_ref"], high["values_ref"]
                    ),
                    "high_label": high["senscase"],
                    "high_tooltip": high["values_ref"],
                    "true_high": high["values"],
                    "high_reals": high["reals"],
                }
            )
        return low_high_per_sensitivity

    def _cut_sensitivities_by_ref(self) -> None:
        """Removes sensitivities smaller than reference sensitivity from table"""

        self._tornadotable = self._tornadotable.loc[
            ((self._tornadotable["low"] - self._tornadotable["high"]) != 0)
            | (self._tornadotable["sensname"] == self._reference)
        ]

    def _sort_sensitivities_by_max(self) -> None:
        """Sorts table based on max(abs('low', 'high'))"""
        self._tornadotable["max"] = (
            self._tornadotable[["low", "high"]]
            .apply(lambda x: max(x.min(), x.max(), key=abs), axis=1)  # type: ignore
            .abs()
        )
        self._tornadotable.sort_values("max", ascending=True, inplace=True)
        self._tornadotable.drop(["max"], axis=1, inplace=True)
        self._tornadotable = pd.concat(
            [
                self._tornadotable[self._tornadotable["sensname"] != self._reference],
                self._tornadotable[self._tornadotable["sensname"] == self._reference],
            ]
        )

    @property
    def low_high_realizations_list(self) -> Dict[str, Dict]:
        return {
            sensname: {
                "real_low": sens_name_df["low_reals"].tolist()[0],
                "real_high": sens_name_df["high_reals"].tolist()[0],
            }
            for sensname, sens_name_df in self.tornadotable.groupby("sensname")
        }

    @staticmethod
    def calc_low_base(low: float, high: float) -> float:
        """
        From the low and high value of a parameter,
        calculates the base (starting x value) of the
        bar visualizing low values.
        """
        if low < 0:
            return min(0, high)
        return low

    @staticmethod
    def calc_high_base(low: float, high: float) -> float:
        """
        From the low and high value of a parameter,
        calculates the base (starting x value) of the bar
        visualizing high values.
        """
        if high > 0:
            return max(0, low)
        return high

    @staticmethod
    def calc_high_x(low: float, high: float) -> float:
        """
        From the low and high value of a parameter,
        calculates the x-value (length of bar) of the bar
        visualizing high values.
        """
        if high > 0:
            base = max(0, low)
            return high - base
        return 0.0

    @staticmethod
    def calc_low_x(low: float, high: float) -> float:
        """
        From the low and high value of a parameter,
        calculates the x-value (length of bar) of the bar
        visualizing low values.
        """
        if low < 0:
            base = min(0, high)
            return low - base
        return 0.0



import json
import sys
from dataclasses import dataclass
from typing import List, Optional, Union

import pandas as pd
import webviz_core_components as wcc
from dash import Input, Output, callback
from dash.development.base_component import Component
from dash.exceptions import PreventUpdate
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC, ViewABC

from webviz_subsurface._components.tornado._tornado_bar_chart import TornadoBarChart
from webviz_subsurface._components.tornado._tornado_data import TornadoData

from ...shared_settings import FilterOption, Scale
from .view_elements import TornadoPlot

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated


class PlotOption(StrEnum):
    SHOW_REAL_POINTS = "show-real-points"
    COLOR_BARS_BY_SENS = "color-bars-by-sens"


class LabelOption(StrEnum):
    SIMPLE = "simple"
    DETAILED = "detailed"
    HIDE = "hide"


class PlotSettings(SettingsGroupABC):
    class IDs(StrEnum):
        PLOT_OPTIONS = "plot-options"
        LABEL = "label"

    def __init__(
        self,
    ) -> None:
        super().__init__("Plot settings")

    def layout(self) -> List[Component]:
        return [
            wcc.Checklist(
                id=self.register_component_unique_id(PlotSettings.IDs.PLOT_OPTIONS),
                label="Plot options",
                options=[
                    {
                        "label": "Show realization points",
                        "value": PlotOption.SHOW_REAL_POINTS,
                    },
                    {
                        "label": "Color bars by sensitivity",
                        "value": PlotOption.COLOR_BARS_BY_SENS,
                    },
                ],
                value=[],
                labelStyle={"display": "block"},
            ),
            wcc.Dropdown(
                id=self.register_component_unique_id(PlotSettings.IDs.LABEL),
                label="Label",
                options=[
                    {"label": "No label", "value": LabelOption.HIDE},
                    {
                        "label": "Simple label",
                        "value": LabelOption.SIMPLE,
                    },
                    {
                        "label": "Detailed label",
                        "value": LabelOption.DETAILED,
                    },
                ],
                value=LabelOption.DETAILED,
                clearable=False,
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.component_unique_id(PlotSettings.IDs.LABEL).to_string(),
                "disabled",
            ),
            Input(
                self.component_unique_id(PlotSettings.IDs.PLOT_OPTIONS).to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _disable_label(plot_options: Optional[List[PlotOption]]) -> bool:
            if plot_options is None:
                return False
            return PlotOption.SHOW_REAL_POINTS in plot_options


class TornadoPlotView(ViewABC):
    class IDs(StrEnum):
        TORNADO_PLOT = "tornado-plot"
        MAIN_COLUMN = "main-column"
        SETTINGS = "settings"

    @dataclass
    class Slots:
        reference: Annotated[Input, str]
        scale: Annotated[Input, str]
        filter_options: Annotated[Input, List[str]]
        data: Annotated[Input, Union[str, bytes, bytearray]]
        sens_filter: Annotated[Input, Union[List[str], str]]

    def __init__(
        self, design_matrix_df: pd.DataFrame, plotly_theme: dict, slots: Slots
    ) -> None:
        super().__init__("Tornado Plot View")

        self._design_matrix_df = design_matrix_df
        self._plotly_theme = plotly_theme
        self._slots = slots

        column = self.add_column(TornadoPlotView.IDs.MAIN_COLUMN)
        first_row = column.make_row()
        first_row.add_view_element(TornadoPlot(), TornadoPlotView.IDs.TORNADO_PLOT)

        self.add_settings_group(PlotSettings(), TornadoPlotView.IDs.SETTINGS)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.view_element(TornadoPlotView.IDs.TORNADO_PLOT)
                .component_unique_id(TornadoPlot.IDs.GRAPH)
                .to_string(),
                "figure",
            ),
            self._slots.reference,
            self._slots.scale,
            self._slots.filter_options,
            Input(
                self.settings_group(TornadoPlotView.IDs.SETTINGS)
                .component_unique_id(PlotSettings.IDs.PLOT_OPTIONS)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(TornadoPlotView.IDs.SETTINGS)
                .component_unique_id(PlotSettings.IDs.LABEL)
                .to_string(),
                "value",
            ),
            self._slots.data,
            self._slots.sens_filter,
        )
        @callback_typecheck
        def _calc_tornado_plot(
            reference: str,
            scale: Scale,
            filter_options: List[FilterOption],
            plot_options: Optional[List[PlotOption]],
            label_option: LabelOption,
            data: Union[str, bytes, bytearray],
            sens_filter: Union[List[str], str],
        ) -> dict:
            if not data:
                raise PreventUpdate
            plot_options = plot_options if plot_options else []
            data = json.loads(data)
            if not isinstance(sens_filter, List):
                sens_filter = [sens_filter]
            if not isinstance(data, dict):
                raise PreventUpdate

            values = pd.DataFrame(data["data"], columns=["REAL", "VALUE"])
            realizations = self._design_matrix_df.loc[
                self._design_matrix_df["ENSEMBLE"] == data["ENSEMBLE"]
            ]

            design_and_responses = pd.merge(values, realizations, on="REAL")
            if sens_filter is not None:
                if reference not in sens_filter:
                    sens_filter.append(reference)
                design_and_responses = design_and_responses.loc[
                    design_and_responses["SENSNAME"].isin(sens_filter)
                ]

            tornado_data = TornadoData(
                dframe=design_and_responses,
                response_name=data.get("response_name"),
                reference=reference,
                scale="Percentage" if scale == Scale.REL_VALUE_PERC else "Absolute",
                cutbyref=FilterOption.REMOVE_SENS_WITH_NO_IMPACT in filter_options,
            )

            tornado_figure = TornadoBarChart(
                tornado_data=tornado_data,
                plotly_theme=self._plotly_theme,
                label_options=label_option,
                number_format=data.get("number_format", ""),
                unit=data.get("unit", ""),
                spaced=data.get("spaced", True),
                locked_si_prefix=data.get("locked_si_prefix", None),
                use_true_base=scale == Scale.TRUE_VALUE,
                show_realization_points=PlotOption.SHOW_REAL_POINTS in plot_options,
                color_by_sensitivity=PlotOption.COLOR_BARS_BY_SENS in plot_options,
            )
            return tornado_figure.figure



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional

import numpy as np

from .colors import hex_to_rgba_str


@dataclass
class FreeLineData:
    """
    Dataclass for defining statistics data for free line trace in fanchart

    `Attributes:`
    * `name` - Name of statistics data
    * `data` - 1D np.array of statistics value data
    """

    name: str
    data: np.ndarray


@dataclass
class LowHighData:
    """
    Definition of paired low and high percentile data for fanchart

    `Attributes:`
    * `low_data`  - 1D np.array of low percentile value data
    * `low_name`  - Name of low percentile
    * `high_data` - 1D np.array of high percentile value data
    * `high_name` - Name of high percentile
    """

    low_data: np.ndarray
    low_name: str
    high_data: np.ndarray
    high_name: str


@dataclass
class MinMaxData:
    """
    Definition of paired minimum and maximum data for fanchart

    `Attributes:`
    * `minimum` - 1D np.array of minimum value data
    * `maximum` - 1D np.array of maximum value data
    """

    minimum: np.ndarray
    maximum: np.ndarray


@dataclass
class FanchartData:
    """
    Dataclass defining fanchart data utilized in creation of statistical fanchart traces

    `Attributes:`
    * `samples` - Common sample point list for each following value list.
    * `free_line` - FreeLineData with name and value data for free line trace in fanchart (e.g.
     mean, median, etc.)
    * `minimum_maximum` - Paired optional minimum and maximum data for fanchart plotting
    * `low_high` - Paired optional low and high percentile names and data for fanchart plotting

    """

    samples: list = field(default_factory=list)
    free_line: Optional[FreeLineData] = None
    minimum_maximum: Optional[MinMaxData] = None
    low_high: Optional[LowHighData] = None


class TraceDirection(Enum):
    HORIZONTAL = 1
    VERTICAL = 2


def validate_fanchart_data(data: FanchartData) -> None:
    """
    Validation of fanchart data

    Ensure equal length of all statistical fanchart data lists and x-axis data list

    Raise ValueError if lengths are unequal
    """
    samples_length = len(data.samples)
    if samples_length <= 0:
        raise ValueError("Empty x-axis data list in FanchartData")
    if data.free_line is not None and samples_length != len(data.free_line.data):
        raise ValueError(
            "Invalid fanchart mean value data length. len(data.samples) != len(free_line.data)"
        )
    if data.minimum_maximum is not None and samples_length != len(
        data.minimum_maximum.minimum
    ):
        raise ValueError(
            "Invalid fanchart minimum value data length. len(data.samples) "
            "!= len(data.minimum_maximum.minimum)"
        )
    if data.minimum_maximum is not None and samples_length != len(
        data.minimum_maximum.maximum
    ):
        raise ValueError(
            "Invalid fanchart maximum value data length. len(data.samples) != "
            "len(data.minimum_maximum.maximum)"
        )
    if data.low_high is not None and samples_length != len(data.low_high.low_data):
        raise ValueError(
            "Invalid fanchart low percentile value data length. len(data.samples) "
            "!= len(data.low_high.low_data)"
        )
    if data.low_high is not None and samples_length != len(data.low_high.high_data):
        raise ValueError(
            "Invalid fanchart high percentile value data length. "
            "len(data.samples) != len(data.low_high.high_data)"
        )


# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
def get_fanchart_traces(
    data: FanchartData,
    hex_color: str,
    legend_group: str,
    legend_name: Optional[str] = None,
    line_shape: str = "linear",
    xaxis: str = "x",
    yaxis: str = "y",
    show_legend: bool = True,
    direction: TraceDirection = TraceDirection.HORIZONTAL,
    show_hoverinfo: bool = True,
    hovertext: str = "",
    hovertemplate: Optional[str] = None,
    hovermode: Optional[str] = None,
    legendrank: Optional[int] = None,
) -> List[Dict[str, Any]]:
    """
    Utility function for creating statistical fanchart traces

    Takes `data` containing data for each statistical feature as input, and creates a list of traces
    for each feature. Plotly plots traces from front to end of the list, thereby the last trace is
    plotted on top.

    Note that min and max, and high and low percentile are paired optional statistics. This implies
    that if minimum is provided, maximum must be provided as well, and vice versa. The same yields
    for low and high percentile data.

    The function provides a list of traces: [trace0, tract1, ..., traceN]

    Fanchart is created by use of fill "tonexty" configuration for the traces. Fill "tonexty" is
    missleading naming, as "tonexty" in trace1 fills to y in trace0, i.e y in previous trace.

    The order of traces are minimum, low, high, maximum and free line. Thus it is required that
    values in minimum <= low, and low <= high, and high <= maximum. Fill is setting "tonexty" in
    this function is set s.t. trace fillings are not stacked making colors in fills unchanged
    when disabling trace statistics inputs (minimum and maximum or low and high).

    Free line is last trace and is plotted on top as a line - without filling to other traces.

    Note:
    If hovertemplate is proved it overrides the hovertext

    Returns:
    List of fanchart traces, one for each statistical feature in data input.
    [trace0, tract1, ..., traceN].
    """

    validate_fanchart_data(data)

    fill_color_light = hex_to_rgba_str(hex_color, 0.3)
    fill_color_dark = hex_to_rgba_str(hex_color, 0.6)
    line_color = hex_to_rgba_str(hex_color, 1)

    def get_default_trace(statistics_name: str, values: np.ndarray) -> Dict[str, Any]:
        trace = {
            "name": legend_name if legend_name else legend_group,
            "x": data.samples if direction == TraceDirection.HORIZONTAL else values,
            "y": values if direction == TraceDirection.HORIZONTAL else data.samples,
            "xaxis": xaxis,
            "yaxis": yaxis,
            "mode": "lines",
            "line": {"width": 0, "color": line_color, "shape": line_shape},
            "legendgroup": legend_group,
            "showlegend": False,
        }
        if legendrank:
            trace["legendrank"] = legendrank
        if not show_hoverinfo:
            trace["hoverinfo"] = "skip"
            return trace
        if hovertemplate is not None:
            trace["hovertemplate"] = hovertemplate + statistics_name
        else:
            trace["hovertext"] = statistics_name + " " + hovertext
        if hovermode is not None:
            trace["hovermode"] = hovermode
        return trace

    traces: List[Dict[str, Any]] = []

    # Minimum
    if data.minimum_maximum is not None:
        traces.append(
            get_default_trace(
                statistics_name="Minimum",
                values=data.minimum_maximum.minimum,
            )
        )

    # Low and high percentile
    if data.low_high is not None:
        low_trace = get_default_trace(
            statistics_name=data.low_high.low_name,
            values=data.low_high.low_data,
        )
        # Add fill to previous trace
        if len(traces) > 0:
            low_trace["fill"] = "tonexty"
            low_trace["fillcolor"] = fill_color_light
        traces.append(low_trace)

        high_trace = get_default_trace(
            statistics_name=data.low_high.high_name,
            values=data.low_high.high_data,
        )
        high_trace["fill"] = "tonexty"
        high_trace["fillcolor"] = fill_color_dark
        traces.append(high_trace)

    # Maximum
    if data.minimum_maximum is not None:
        maximum_trace = get_default_trace(
            statistics_name="Maximum",
            values=data.minimum_maximum.maximum,
        )
        # Add fill to previous trace
        if len(traces) > 0:
            maximum_trace["fill"] = "tonexty"
            maximum_trace["fillcolor"] = fill_color_light
        traces.append(maximum_trace)

    # Free line
    if data.free_line is not None:
        line_trace = get_default_trace(
            statistics_name=data.free_line.name,
            values=data.free_line.data,
        )
        # Set solid line for mean
        line_trace["line"] = {"color": line_color, "shape": line_shape}
        traces.append(line_trace)

    # Set legend for last trace in list
    if len(traces) > 0:
        traces[-1]["showlegend"] = show_legend

    return traces



import datetime
import os
from pathlib import Path
from typing import Optional

import pandas as pd

# The fmu.ensemble dependency ecl is only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    from fmu.ensemble import ScratchEnsemble
except ImportError:
    pass

from webviz_subsurface._providers import (
    EnsembleSummaryProviderFactory,
    Frequency,
    VectorMetadata,
)


# Helper function for generating per-realization CSV files based on aggregated CSV file
def _split_aggr_csv_into_per_real(aggr_csvfile: str, output_folder: str) -> None:
    df = pd.read_csv(aggr_csvfile)
    df = df[df["ENSEMBLE"] == "iter-0"]
    df = df.drop(columns="ENSEMBLE")

    for real in range(0, 10):
        real_df = df[df["REAL"] == real]
        real_df = real_df.drop(columns="REAL")
        os.makedirs(f"{output_folder}/realization-{real}/iter-0/", exist_ok=True)
        real_df.to_csv(
            f"{output_folder}/realization-{real}/iter-0/smry.csv", index=False
        )


# Helper function dumping values obtained via fmu to CSV file.
# Used to find the expected values in "arrow" tests
def _dump_smry_to_csv_using_fmu(
    ens_path: str, time_index: str, output_csv_file: str
) -> None:
    scratch_ensemble = ScratchEnsemble("tempEnsName", paths=ens_path)
    df = scratch_ensemble.load_smry(time_index=time_index)
    df.sort_values(["DATE", "REAL"], inplace=True)

    print("Dataframe shape::", df.shape)

    unique_dates = df["DATE"].unique()
    print("Num unique dates:", len(unique_dates))
    print(unique_dates)

    unique_reals = df["REAL"].unique()
    print("Num unique reals:", len(unique_reals))
    print(unique_reals)

    df.to_csv(output_csv_file, index=False)


def test_create_from_arrow_unsmry_lazy(testdata_folder: Path, tmp_path: Path) -> None:

    ensemble_path = str(testdata_folder / "01_drogon_ahm/realization-*/iter-0")

    # Used to generate test results
    # _dump_smry_to_csv_using_fmu(ensemble_path, "monthly", "expected_smry.csv")

    factory = EnsembleSummaryProviderFactory(tmp_path, allow_storage_writes=True)
    provider = factory.create_from_arrow_unsmry_lazy(
        ens_path=ensemble_path, rel_file_pattern="share/results/unsmry/*.arrow"
    )

    assert provider.supports_resampling()

    assert provider.vector_metadata("FOPT") is not None

    vecnames = provider.vector_names()
    assert len(vecnames) == 931

    dates = provider.dates(Frequency.MONTHLY)
    assert len(dates) == 31
    assert isinstance(dates[0], datetime.datetime)
    assert dates[0] == datetime.datetime(2018, 1, 1)
    assert dates[-1] == datetime.datetime(2020, 7, 1)

    realizations = provider.realizations()
    assert len(realizations) == 100
    assert realizations[0] == 0
    assert realizations[-1] == 99

    vecdf = provider.get_vectors_df(["FOPR"], Frequency.MONTHLY)
    assert vecdf.shape == (3100, 3)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "FOPR"]
    assert vecdf["DATE"].nunique() == 31
    assert vecdf["REAL"].nunique() == 100
    sampleddate = vecdf["DATE"][0]
    assert isinstance(sampleddate, datetime.datetime)

    vecdf = provider.get_vectors_df(["FOPR"], Frequency.MONTHLY, [5])
    assert vecdf.shape == (31, 3)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "FOPR"]
    assert vecdf["DATE"].nunique() == 31
    assert vecdf["REAL"].nunique() == 1
    assert vecdf["REAL"][0] == 5


def test_arrow_unsmry_lazy_vector_metadata(
    testdata_folder: Path, tmp_path: Path
) -> None:

    ensemble_path = str(testdata_folder / "01_drogon_ahm/realization-*/iter-0")
    factory = EnsembleSummaryProviderFactory(tmp_path, allow_storage_writes=True)
    provider = factory.create_from_arrow_unsmry_lazy(
        ens_path=ensemble_path, rel_file_pattern="share/results/unsmry/*.arrow"
    )

    meta: Optional[VectorMetadata] = provider.vector_metadata("FOPR")
    assert meta is not None
    assert meta.unit == "SM3/DAY"
    assert meta.is_total is False
    assert meta.is_rate is True
    assert meta.is_historical is False
    assert meta.keyword == "FOPR"
    assert meta.wgname is None
    assert meta.get_num == 0

    meta = provider.vector_metadata("WOPTH:A6")
    assert meta is not None
    assert meta.unit == "SM3"
    assert meta.is_total is True
    assert meta.is_rate is False
    assert meta.is_historical is True
    assert meta.keyword == "WOPTH"
    assert meta.wgname == "A6"
    assert meta.get_num == 11

    meta = provider.vector_metadata("FWCT")
    assert meta is not None
    assert meta.unit == ""
    assert meta.is_total is False
    assert meta.is_rate is True
    assert meta.is_historical is False
    assert meta.keyword == "FWCT"
    assert meta.wgname is None
    assert meta.get_num == 0


def test_create_from_arrow_unsmry_presampled_monthly(
    testdata_folder: Path, tmp_path: Path
) -> None:

    ensemble_path = testdata_folder / "01_drogon_ahm/realization-*/iter-0"

    factory = EnsembleSummaryProviderFactory(tmp_path, allow_storage_writes=True)
    provider = factory.create_from_arrow_unsmry_presampled(
        ens_path=str(ensemble_path),
        rel_file_pattern="share/results/unsmry/*.arrow",
        sampling_frequency=Frequency.MONTHLY,
    )

    assert not provider.supports_resampling()

    assert provider.vector_metadata("FOPT") is not None

    vecnames = provider.vector_names()
    assert len(vecnames) == 931

    dates = provider.dates(None)
    assert len(dates) == 31
    assert isinstance(dates[0], datetime.datetime)
    assert dates[0] == datetime.datetime(2018, 1, 1)
    assert dates[-1] == datetime.datetime(2020, 7, 1)

    realizations = provider.realizations()
    assert len(realizations) == 100
    assert realizations[0] == 0
    assert realizations[-1] == 99

    vecdf = provider.get_vectors_df(["FOPR"], None)
    assert vecdf.shape == (3100, 3)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "FOPR"]
    assert vecdf["DATE"].nunique() == 31
    assert vecdf["REAL"].nunique() == 100
    sampleddate = vecdf["DATE"][0]
    assert isinstance(sampleddate, datetime.datetime)


def test_create_from_per_realization_csv_file(
    testdata_folder: Path, tmp_path: Path
) -> None:

    _split_aggr_csv_into_per_real(
        str(testdata_folder / "reek_test_data/aggregated_data/smry.csv"),
        str(tmp_path / "fake_data"),
    )

    factory = EnsembleSummaryProviderFactory(tmp_path, allow_storage_writes=True)

    ens_path = tmp_path / "fake_data/realization-*/iter-0"
    csvfile = "smry.csv"
    provider = factory.create_from_per_realization_csv_file(str(ens_path), csvfile)

    vecnames = provider.vector_names()
    assert len(vecnames) == 16
    assert vecnames[0] == "FGIP"
    assert vecnames[15] == "YEARS"

    realizations = provider.realizations()
    assert len(realizations) == 10

    vecdf = provider.get_vectors_df(["FOPR"], None)
    assert vecdf.shape == (380, 3)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "FOPR"]
    assert vecdf["REAL"].nunique() == 10
    sampleddate = vecdf["DATE"][0]
    assert isinstance(sampleddate, datetime.datetime)

    vecdf = provider.get_vectors_df(["FOPR"], None, [1])
    assert vecdf.shape == (38, 3)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "FOPR"]
    assert vecdf["REAL"].nunique() == 1


def test_create_from_ensemble_csv(testdata_folder: Path, tmp_path: Path) -> None:

    factory = EnsembleSummaryProviderFactory(tmp_path, allow_storage_writes=True)

    csv_filename = (
        testdata_folder / "reek_test_data" / "aggregated_data" / "smry_hm.csv"
    )
    provider = factory.create_from_ensemble_csv_file(csv_filename, "iter-0")

    vecnames = provider.vector_names()
    assert len(vecnames) == 473
    assert vecnames[0] == "BPR:15,28,1"
    assert vecnames[472] == "YEARS"

    realizations = provider.realizations()
    assert len(realizations) == 10

    dates = provider.dates(None)
    assert len(dates) == 38
    assert isinstance(dates[0], datetime.datetime)
    assert dates[0] == datetime.datetime(2000, 1, 1)
    assert dates[-1] == datetime.datetime(2003, 2, 1)

    vecdf = provider.get_vectors_df(["FOPR"], None)
    assert vecdf.shape == (380, 3)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "FOPR"]
    assert vecdf["REAL"].nunique() == 10
    sampleddate = vecdf["DATE"][0]
    assert isinstance(sampleddate, datetime.datetime)

    vecdf = provider.get_vectors_df(["FOPR"], None, [1])
    assert vecdf.shape == (38, 3)
    assert vecdf.columns.tolist() == ["DATE", "REAL", "FOPR"]
    assert vecdf["REAL"].nunique() == 1



import base64
import io
from typing import Optional

import numpy as np
import xtgeo
from PIL import Image

from webviz_subsurface._datainput.image_processing import array2d_to_png


class SurfaceLeafletModel:
    """Class to make a wcc.LeafletMap layer from a Xtgeo RegularSurface"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        surface: xtgeo.RegularSurface,
        name: Optional[str] = None,
        clip_min: Optional[float] = None,
        clip_max: Optional[float] = None,
        unit: str = " ",
        apply_shading: bool = False,
        colors: Optional[list] = None,
        updatemode: str = "update",
    ):
        self.name = name if name is not None else surface.name
        self.surface = surface
        self.apply_shading = apply_shading
        self.updatemode = updatemode
        self.bounds = [[surface.xmin, surface.ymin], [surface.xmax, surface.ymax]]
        self.zvalues = self.get_zvalues(clip_min=clip_min, clip_max=clip_max)
        self.unit = unit
        self.colors = self.set_colors(colors)

    @property
    def img_url(self) -> str:
        return array2d_to_png(self.scaled_zvalues.copy())

    @property
    def min_val(self) -> float:
        return np.nanmin(self.zvalues)

    @property
    def max_val(self) -> float:
        return np.nanmax(self.zvalues)

    @property
    def scale_factor(self) -> float:
        if self.min_val == 0.0 and self.max_val == 0.0:
            return 1.0
        return (256 * 256 * 256 - 1) / (self.max_val - self.min_val)

    @property
    def map_scale(self) -> float:
        img = Image.open(
            io.BytesIO(
                base64.b64decode(array2d_to_png(self.scaled_zvalues.copy())[22:])
            )
        )
        width, height = img.size
        if width * height >= 300 * 300:
            return 1.0
        ratio = (1000**2) / (width * height)
        return np.sqrt(ratio).round(2)

    def get_zvalues(
        self,
        unrotate: bool = True,
        flip: bool = True,
        clip_min: float = None,
        clip_max: float = None,
    ) -> np.ndarray:
        surface = self.surface.copy()
        if clip_min or clip_max:
            np.ma.clip(surface.values, clip_min, clip_max, out=surface.values)  # type: ignore
        if unrotate:
            surface.unrotate()
        surface.fill(np.nan)
        values = surface.values
        if flip:
            values = np.flip(values.transpose(), axis=0)
        # If all values are masked set to zero
        if values.mask.all():
            values = np.zeros(values.shape)
        return values

    @property
    def scaled_zvalues(self) -> np.ndarray:
        return (self.zvalues - self.min_val) * self.scale_factor

    @staticmethod
    def set_colors(colors: list = None) -> list:
        return (
            [
                "#fde725",
                "#b5de2b",
                "#6ece58",
                "#35b779",
                "#1f9e89",
                "#26828e",
                "#31688e",
                "#3e4989",
                "#482878",
                "#440154",
            ]
            if colors is None
            else colors
        )

    @property
    def layer(self) -> dict:
        return {
            "name": self.name,
            "checked": True,
            "id": self.name,
            "action": self.updatemode,
            "baseLayer": True,
            "data": [
                {
                    "type": "image",
                    "url": self.img_url,
                    "colorScale": {
                        "colors": self.colors,
                        "scaleType": "linear",
                    },
                    "shader": {
                        "applyHillshading": self.apply_shading,
                    },
                    "bounds": self.bounds,
                    "minvalue": self.min_val,
                    "maxvalue": self.max_val,
                    "unit": self.unit,
                    "imageScale": self.map_scale,
                }
            ],
        }



from typing import Any, Dict, List, Optional, Tuple

import webviz_core_components as wcc
import webviz_subsurface_components as wsc
from dash import Input, Output, State, callback, html
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC, ViewABC

from ..._utils import WellCompletionDataModel
from ._view_element import WellCompletionViewElement


class DataMode(StrEnum):
    AGGREGATED = "aggregated"
    SINGLE_REAL = "single-real"


class ViewSettings(SettingsGroupABC):
    class Ids(StrEnum):
        ENSEMBLE = "ensemble"
        DATA_MODE = "mode"
        REALIZATION = "realization"
        REAL_BLOCK = "real-block"

    def __init__(self, ensembles: List[str]) -> None:
        super().__init__("Settings")
        self._ensembles = ensembles

    def layout(self) -> List[Any]:
        return [
            wcc.Dropdown(
                id=self.register_component_unique_id(ViewSettings.Ids.ENSEMBLE),
                label="Ensemble",
                options=[{"label": ens, "value": ens} for ens in self._ensembles],
                clearable=False,
                value=self._ensembles[0],
                persistence=True,
                persistence_type="session",
            ),
            wcc.RadioItems(
                id=self.register_component_unique_id(ViewSettings.Ids.DATA_MODE),
                label="Aggregated or single realization",
                options=[
                    {
                        "label": "Aggregated",
                        "value": DataMode.AGGREGATED,
                    },
                    {
                        "label": "Single realization",
                        "value": DataMode.SINGLE_REAL,
                    },
                ],
                value=DataMode.AGGREGATED,
            ),
            html.Div(
                id=self.register_component_unique_id(ViewSettings.Ids.REAL_BLOCK),
                children=wcc.Dropdown(
                    id=self.register_component_unique_id(ViewSettings.Ids.REALIZATION),
                    label="Realization",
                    options=[],
                    multi=False,
                ),
            ),
        ]


class WellCompletionView(ViewABC):
    class Ids(StrEnum):
        VIEW_ELEMENT = "view-element"
        SETTINGS = "settings"

    def __init__(self, data_models: Dict[str, WellCompletionDataModel]) -> None:
        super().__init__("Well Completion")

        self._data_models = data_models

        self.add_settings_group(
            ViewSettings(list(self._data_models.keys())),
            WellCompletionView.Ids.SETTINGS,
        )

        column = self.add_column()
        column.add_view_element(
            WellCompletionViewElement(), WellCompletionView.Ids.VIEW_ELEMENT
        )

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.view_element(WellCompletionView.Ids.VIEW_ELEMENT)
                .component_unique_id(WellCompletionViewElement.Ids.COMPONENT)
                .to_string(),
                "children",
            ),
            Output(
                self.view_element(WellCompletionView.Ids.VIEW_ELEMENT)
                .component_unique_id(WellCompletionViewElement.Ids.COMPONENT)
                .to_string(),
                "style",
            ),
            Input(
                self.settings_group(WellCompletionView.Ids.SETTINGS)
                .component_unique_id(ViewSettings.Ids.ENSEMBLE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(WellCompletionView.Ids.SETTINGS)
                .component_unique_id(ViewSettings.Ids.DATA_MODE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(WellCompletionView.Ids.SETTINGS)
                .component_unique_id(ViewSettings.Ids.REALIZATION)
                .to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _render_well_completions(
            ensemble_name: str, data_mode: DataMode, real: int
        ) -> Tuple:

            data = self._data_models[ensemble_name].create_ensemble_dataset(
                realization=real if data_mode == DataMode.SINGLE_REAL else None
            )
            no_leaves = _count_leaves(data["stratigraphy"])
            return (
                wsc.WellCompletions(id="well-completion-component", data=data),
                {
                    "height": no_leaves * 50 + 180,
                    "min-height": 500,
                    "width": "98%",
                },
            )

        @callback(
            Output(
                self.settings_group(WellCompletionView.Ids.SETTINGS)
                .component_unique_id(ViewSettings.Ids.REALIZATION)
                .to_string(),
                "options",
            ),
            Output(
                self.settings_group(WellCompletionView.Ids.SETTINGS)
                .component_unique_id(ViewSettings.Ids.REALIZATION)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(WellCompletionView.Ids.SETTINGS)
                .component_unique_id(ViewSettings.Ids.ENSEMBLE)
                .to_string(),
                "value",
            ),
            State(
                self.settings_group(WellCompletionView.Ids.SETTINGS)
                .component_unique_id(ViewSettings.Ids.REALIZATION)
                .to_string(),
                "value",
            ),
        )
        def _update_realization_dropdown(
            ensemble: str, state_real: int
        ) -> Tuple[List[Dict[str, Any]], Optional[int]]:
            """Updates the ealization dropdown with ensemble values"""
            reals = self._data_models[ensemble].realizations
            return (
                [{"label": real, "value": real} for real in reals],
                state_real if state_real in reals else reals[0],
            )

        @callback(
            Output(
                self.settings_group(WellCompletionView.Ids.SETTINGS)
                .component_unique_id(ViewSettings.Ids.REAL_BLOCK)
                .to_string(),
                component_property="style",
            ),
            Input(
                self.settings_group(WellCompletionView.Ids.SETTINGS)
                .component_unique_id(ViewSettings.Ids.DATA_MODE)
                .to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _show_hide_single_real_options(data_mode: DataMode) -> Dict[str, str]:
            """Hides or unhides the realization dropdown according to whether mean
            or single realization is selected.
            """
            if data_mode == DataMode.AGGREGATED:
                return {"display": "none"}
            return {"display": "block"}


def _count_leaves(stratigraphy: List[Dict[str, Any]]) -> int:
    """Counts the number of leaves in the stratigraphy tree"""
    return sum(
        _count_leaves(zonedict["subzones"]) if "subzones" in zonedict else 1
        for zonedict in stratigraphy
    )



from typing import Callable

import pandas as pd
from dash import ALL, Input, Output, State, callback, callback_context, dcc
from dash.exceptions import PreventUpdate


def export_data_controllers(get_uuid: Callable) -> None:
    @callback(
        Output(get_uuid("download-dataframe"), "data"),
        Input({"request": "table_data", "table_id": ALL}, "data_requested"),
        State({"table_id": ALL}, "data"),
        State({"table_id": ALL}, "columns"),
        State({"request": "table_data", "table_id": ALL}, "id"),
        State({"table_id": ALL}, "id"),
        prevent_initial_call=True,
    )
    def _export_table_data(
        _data_requested: list,
        table_data: list,
        table_columns: list,
        button_ids: list,
        table_ids: list,
    ) -> Callable:

        ctx = callback_context.triggered[0]
        export_clicks = {
            id_value["table_id"]: n_clicks
            for id_value, n_clicks in zip(button_ids, _data_requested)
        }
        table_to_extract = [x for x in export_clicks.keys() if x in ctx["prop_id"]]
        if not table_to_extract or export_clicks[table_to_extract[0]] is None:
            raise PreventUpdate

        index = [x["table_id"] for x in table_ids].index(table_to_extract[0])
        table_data = table_data[index]
        table_columns = [x["name"] for x in table_columns[index]]

        return dcc.send_data_frame(
            pd.DataFrame(data=table_data, columns=table_columns).to_excel,
            "VolumetricAnalysis.xlsx",
            index=False,
        )



from typing import Dict, List

import pandas as pd
import webviz_core_components as wcc
from dash import Input, Output, callback, html
from dash.development.base_component import Component
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from .._utils._plot_utils import ColorBy


class DataSettings(SettingsGroupABC):

    PHASES = ["OIL", "GAS", "WATER"]
    phases_additional_info: List[str] = []

    class Ids(StrEnum):
        COLOR_BY = "select-color"
        ENSEMBLES = "ensembles"
        PHASE = "phase"
        PVTNUM = "pvtnum"
        PVTNUM_BOX = "pvtnum-box"
        ENSEMBLES_BOX = "ensembles-box"

    def __init__(self, pvt_data_frame: pd.DataFrame) -> None:
        super().__init__("Data")

        self.pvt_data_frame = pvt_data_frame

        self.ensembles = list(self.pvt_data_frame["ENSEMBLE"].unique())

        self.pvtnum = list(self.pvt_data_frame["PVTNUM"].unique())

        if self.pvt_data_frame["KEYWORD"].str.contains("PVTO").any():
            DataSettings.phases_additional_info.append("PVTO")
        elif self.pvt_data_frame["KEYWORD"].str.contains("PVDO").any():
            DataSettings.phases_additional_info.append("PVDO")
        elif self.pvt_data_frame["KEYWORD"].str.contains("PVCDO").any():
            DataSettings.phases_additional_info.append("PVCDO")
        if self.pvt_data_frame["KEYWORD"].str.contains("PVTG").any():
            DataSettings.phases_additional_info.append("PVTG")
        elif self.pvt_data_frame["KEYWORD"].str.contains("PVDG").any():
            DataSettings.phases_additional_info.append("PVDG")
        if self.pvt_data_frame["KEYWORD"].str.contains("PVTW").any():
            DataSettings.phases_additional_info.append("PVTW")

        self._ensemble_properties: dict = {}
        self._pvtnum_properties: dict = {}

    @property
    def phases(self) -> Dict[str, str]:
        phase_descriptions: Dict[str, str] = {}
        for i, phase in enumerate(DataSettings.PHASES):
            phase_descriptions[phase] = self.phases_additional_info[i]
        return phase_descriptions

    def layout(self) -> List[Component]:
        self._ensemble_properties = dict(
            id=self.register_component_unique_id(DataSettings.Ids.ENSEMBLES),
            label="Ensembles",
            options=[{"label": x, "value": x} for x in self.ensembles],
            vertical=True,
        )
        self._pvtnum_properties = dict(
            id=self.register_component_unique_id(DataSettings.Ids.PVTNUM),
            label="Pvtnum",
            options=[{"label": x, "value": x} for x in self.pvtnum],
            vertical=True,
        )

        return [
            wcc.RadioItems(
                id=self.register_component_unique_id(DataSettings.Ids.COLOR_BY),
                label="Color by",
                options=[{"label": x, "value": x} for x in ColorBy],
                value=ColorBy.ENSEMBLE,
                vertical=True,
            ),
            html.Div(
                id=self.register_component_unique_id(DataSettings.Ids.ENSEMBLES_BOX),
                children=[
                    wcc.Checklist(**self._ensemble_properties, value=self.ensembles)
                ],
            ),
            wcc.Dropdown(
                id=self.register_component_unique_id(DataSettings.Ids.PHASE),
                label="Phase",
                options=[
                    {
                        "label": f"{value.lower().capitalize()} ({info})",
                        "value": value,
                    }
                    for value, info in self.phases.items()
                ],
                value=list(self.phases.items())[0][0],
                clearable=False,
                persistence=False,
            ),
            html.Div(
                id=self.register_component_unique_id(DataSettings.Ids.PVTNUM_BOX),
                children=[
                    wcc.RadioItems(**self._pvtnum_properties, value=self.pvtnum[0])
                ],
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.component_unique_id(DataSettings.Ids.ENSEMBLES_BOX).to_string(),
                "children",
            ),
            Output(
                self.component_unique_id(DataSettings.Ids.PVTNUM_BOX).to_string(),
                "children",
            ),
            Input(
                self.component_unique_id(DataSettings.Ids.COLOR_BY).to_string(), "value"
            ),
        )
        @callback_typecheck
        def _update_ensembles_pvtnum(selected_color_by: ColorBy) -> List[Component]:
            if selected_color_by is ColorBy.ENSEMBLE:
                return [
                    wcc.Checklist(
                        **self._ensemble_properties,
                        value=self.ensembles,
                    ),
                    wcc.RadioItems(
                        **self._pvtnum_properties,
                        value=self.pvtnum[0],
                    ),
                ]
            return [
                wcc.RadioItems(
                    **self._ensemble_properties,
                    value=self.ensembles[0],
                ),
                wcc.Checklist(
                    **self._pvtnum_properties,
                    value=self.pvtnum,
                ),
            ]



import glob
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional

from fmu.ensemble import ScratchEnsemble


@dataclass(frozen=True)
class EclipseCaseFileInfo:
    realization: int
    egrid_path: str
    init_path: str
    unrst_path: str


def _discover_ensemble_realizations_fmu(ens_path: str) -> Dict[int, str]:
    """Returns dict indexed by realization number and with runpath as value"""
    scratch_ensemble = ScratchEnsemble("dummyEnsembleName", paths=ens_path).filter("OK")
    real_dict = {i: r.runpath() for i, r in scratch_ensemble.realizations.items()}
    return real_dict


def _discover_ensemble_realizations(ens_path: str) -> Dict[int, str]:
    # Much faster than FMU impl above, but is it risky?
    # Do we need to check for OK-file?
    real_dict: Dict[int, str] = {}

    realidxregexp = re.compile(r"realization-(\d+)")
    globbed_real_dirs = sorted(glob.glob(str(ens_path)))
    for real_dir in globbed_real_dirs:
        realnum: Optional[int] = None
        for path_comp in reversed(real_dir.split(os.path.sep)):
            realmatch = re.match(realidxregexp, path_comp)
            if realmatch:
                realnum = int(realmatch.group(1))
                break

        if realnum is not None:
            real_dict[realnum] = real_dir

    return real_dict


def discover_per_realization_eclipse_files(
    ens_path: str, grid_name: str
) -> List[EclipseCaseFileInfo]:
    rel_folder: str = "eclipse/model"

    real_dict = _discover_ensemble_realizations_fmu(ens_path)
    eclipse_file_infos = []
    for realnum, runpath in sorted(real_dict.items()):
        globbed_model_folder = glob.glob(str(Path(runpath) / rel_folder))

        for folder in sorted(globbed_model_folder):
            egrid_file = f"{folder}/{grid_name}.EGRID"
            init_file = f"{folder}/{grid_name}.INIT"
            unrst_file = f"{folder}/{grid_name}.UNRST"
            if (
                not os.path.exists(egrid_file)
                or not os.path.exists(init_file)
                or not os.path.exists(unrst_file)
            ):
                continue
            eclipse_file_infos.append(
                EclipseCaseFileInfo(
                    realization=realnum,
                    egrid_path=egrid_file,
                    init_path=init_file,
                    unrst_path=unrst_file,
                )
            )
    if not eclipse_file_infos:
        raise ValueError(
            f"No eclipse models found at {Path(ens_path) / rel_folder / grid_name}"
        )
    return eclipse_file_infos



import warnings
from pathlib import Path
from typing import List
from uuid import uuid4

import pandas as pd
import webviz_core_components as wcc
import xtgeo
from dash import Input, Output, State, html
from dash.exceptions import PreventUpdate
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.utils import calculate_slider_step
from webviz_config.webviz_store import webvizstore
from webviz_subsurface_components import LeafletMap

from webviz_subsurface._models import SurfaceLeafletModel

from .._datainput.grid import load_grid, load_grid_parameter
from .._datainput.surface import get_surface_fence


class SurfaceWithGridCrossSection(WebvizPluginABC):
    """Visualizes surfaces in a map view and grid parameters in a cross section view. \
The cross section is defined by a polyline interactively edited in the map view.

!> This is an experimental plugin exploring how we can visualize 3D grid data in Webviz. \
The performance is currently slow for large grids.

---

* **`gridfile`:** Path to grid geometry (`ROFF` format) (absolute or relative to config file).
* **`gridparameterfiles`:** List of file paths to grid parameters (`ROFF` format) \
 (absolute or relative to config file).
* **`gridparameternames`:** List corresponding to filepaths of displayed parameter names.
* **`surfacefiles`:** List of file paths to surfaces (`irap binary` format) \
 (absolute or relative to config file).
* **`surfacenames`:** List corresponding to file paths of displayed surface names.
* **`zunit`:** z-unit for display.
* **`colors`:** List of hex colors to use. \
Note that apostrophies should be used to avoid that hex colors are read as comments. E.g. \
`'#000000'` for black.

---
**Example files**

* [Gridfile](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
reek_history_match/realization-0/iter-0/share/results/grids/geogrid.roff).
* [One file for gridparameterfiles](https://github.com/equinor/webviz-subsurface-testdata/blob/\
master/reek_history_match/realization-0/iter-0/share/results/grids/geogrid--poro.roff).
* [One file for surfacefiles](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
reek_history_match/realization-0/iter-0/share/results/\
maps/topupperreek--ds_extracted_horizons.gri).

The files above are on a `ROFF binary` format and can be investigated outside `webviz` using \
e.g. [xtgeo](https://xtgeo.readthedocs.io/en/latest/).
"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app,
        webviz_settings: WebvizSettings,
        gridfile: Path,
        gridparameterfiles: List[Path],
        surfacefiles: List[Path],
        gridparameternames: list = None,
        surfacenames: list = None,
        zunit="depth (m)",
        colors: list = None,
    ):

        super().__init__()
        self.zunit = zunit
        self.gridfile = str(gridfile)
        self.gridparafiles = [str(gridfile) for gridfile in gridparameterfiles]
        self.surfacefiles = [str(surffile) for surffile in surfacefiles]
        if surfacenames is not None:
            if len(surfacenames) != len(surfacefiles):
                raise ValueError(
                    "List of surface names specified should be same length as list of surfacefiles"
                )
            self.surfacenames = surfacenames
        else:
            self.surfacenames = [Path(surfacefile).stem for surfacefile in surfacefiles]
        if gridparameternames is not None:
            if len(gridparameternames) != len(gridparameterfiles):
                raise ValueError(
                    "List of grid parameter names specified should be same length "
                    "as list of gridparameterfiles"
                )
            self.gridparanames = gridparameternames
        else:
            self.gridparanames = [
                Path(gridfile).stem for gridfile in gridparameterfiles
            ]
        self.plotly_theme = webviz_settings.theme.plotly_theme
        self.initial_colors = (
            colors
            if colors is not None
            else [
                "#440154",
                "#482878",
                "#3e4989",
                "#31688e",
                "#26828e",
                "#1f9e89",
                "#35b779",
                "#6ece58",
                "#b5de2b",
                "#fde725",
            ]
        )
        self.uid = uuid4()
        self.set_callbacks(app)

    def ids(self, element):
        """Generate unique id for dom element"""
        return f"{element}-id-{self.uid}"

    @property
    def tour_steps(self):
        return [
            {
                "id": self.ids("layout"),
                "content": (
                    "Plugin to display surfaces and random lines from a 3D grid parameter. "
                ),
            },
            {
                "id": self.ids("surface"),
                "content": ("The visualized surface."),
            },
            {
                "id": self.ids("map-view"),
                "content": (
                    "Map view of the surface. Use the right toolbar to "
                    "draw a random line."
                ),
            },
            {
                "id": self.ids("fence-view"),
                "content": (
                    "Cross section view of the grid parameter along the edited line. "
                    "The view is empty until a random line is drawn in the map view."
                ),
            },
            {
                "id": self.ids("surface-type"),
                "content": (
                    "Display the z-value of the surface (e.g. depth) or "
                    "the grid parameter value where the surface intersect the grid."
                ),
            },
            {
                "id": self.ids("gridparameter"),
                "content": "The visualized grid parameter.",
            },
            {
                "id": self.ids("color-scale"),
                "content": ("Click this button to change colorscale"),
            },
            {
                "id": self.ids("color-values"),
                "content": ("Drag either node of slider to truncate color ranges"),
            },
            {
                "id": self.ids("color-range-btn"),
                "content": (
                    "Click this button to update color slider min/max and reset ranges."
                ),
            },
        ]

    @property
    def layout(self):
        return wcc.FlexBox(
            id=self.ids("layout"),
            children=[
                wcc.Frame(
                    style={"flex": 1},
                    children=[
                        wcc.Selectors(
                            label="Map settings",
                            children=[
                                wcc.Dropdown(
                                    id=self.ids("surface"),
                                    label="Select surface",
                                    options=[
                                        {"label": name, "value": path}
                                        for name, path in zip(
                                            self.surfacenames, self.surfacefiles
                                        )
                                    ],
                                    value=self.surfacefiles[0],
                                    clearable=False,
                                ),
                                wcc.RadioItems(
                                    id=self.ids("surface-type"),
                                    options=[
                                        {
                                            "label": "Display surface z-value",
                                            "value": "surface",
                                        },
                                        {
                                            "label": "Display seismic attribute as z-value",
                                            "value": "attribute",
                                        },
                                    ],
                                    value="surface",
                                ),
                            ],
                        ),
                        wcc.Selectors(
                            label="Intersection settings",
                            children=[
                                wcc.Dropdown(
                                    label="Select grid parameter",
                                    id=self.ids("gridparameter"),
                                    options=[
                                        {"label": name, "value": para}
                                        for name, para in zip(
                                            self.gridparanames, self.gridparafiles
                                        )
                                    ],
                                    value=self.gridparafiles[0],
                                    clearable=False,
                                ),
                                wcc.Label(
                                    children="Set colorscale",
                                ),
                                wcc.ColorScales(
                                    id=self.ids("color-scale"),
                                    colorscale=self.initial_colors,
                                    nSwatches=12,
                                ),
                                wcc.RangeSlider(
                                    label="Set color range",
                                    id=self.ids("color-values"),
                                    tooltip={
                                        "placement": "bottom",
                                        "always_visible": True,
                                    },
                                    marks=None,
                                ),
                                html.Button(
                                    id=self.ids("color-range-btn"),
                                    children="Reset Range",
                                ),
                            ],
                        ),
                    ],
                ),
                wcc.Frame(
                    highlight=False,
                    style={
                        "height": "800px",
                        "flex": 3,
                    },
                    children=[
                        LeafletMap(
                            id=self.ids("map-view"),
                            autoScaleMap=True,
                            minZoom=-19,
                            updateMode="update",
                            layers=[],
                            drawTools={
                                "drawMarker": False,
                                "drawPolygon": False,
                                "drawPolyline": True,
                                "position": "topright",
                            },
                            mouseCoords={"position": "bottomright"},
                            colorBar={"position": "bottomleft"},
                            switch={
                                "value": False,
                                "disabled": False,
                                "label": "Hillshading",
                            },
                        ),
                    ],
                ),
                html.Div(
                    style={"flex": 3, "height": "800px"},
                    children=[
                        wcc.Graph(
                            config={"displayModeBar": False}, id=self.ids("fence-view")
                        ),
                    ],
                ),
            ],
        )

    def set_callbacks(self, app):
        @app.callback(
            Output(self.ids("map-view"), "layers"),
            [
                Input(self.ids("surface"), "value"),
                Input(self.ids("surface-type"), "value"),
                Input(self.ids("gridparameter"), "value"),
                Input(self.ids("color-values"), "value"),
                Input(self.ids("map-view"), "switch"),
            ],
        )
        def _render_surface(
            surfacepath,
            surface_type,
            gridparameter,
            color_values,
            hillshade,
        ):

            surface = xtgeo.surface_from_file(get_path(surfacepath))
            min_val = None
            max_val = None

            if surface_type == "attribute":
                min_val = color_values[0] if color_values else None
                max_val = color_values[1] if color_values else None
                grid = load_grid(get_path(self.gridfile))
                gridparameter = load_grid_parameter(grid, get_path(gridparameter))
                surface.slice_grid3d(grid, gridparameter)

            return [
                SurfaceLeafletModel(
                    surface,
                    name="surface",
                    clip_min=min_val,
                    clip_max=max_val,
                    apply_shading=hillshade.get("value", False),
                ).layer
            ]

        @app.callback(
            Output(self.ids("fence-view"), "figure"),
            [
                Input(self.ids("map-view"), "polyline_points"),
                Input(self.ids("gridparameter"), "value"),
                Input(self.ids("surface"), "value"),
                Input(self.ids("color-values"), "value"),
                Input(self.ids("color-scale"), "colorscale"),
            ],
        )
        def _render_fence(coords, gridparameter, surfacepath, color_values, colorscale):
            if not coords:
                raise PreventUpdate
            grid = load_grid(get_path(self.gridfile))
            if (
                grid.subgrids is not None
                and sum([len(subgrid) for subgrid in grid.subgrids.values()])
                != grid.nlay
            ):
                warnings.warn(
                    (
                        f"Subgrid information in {self.gridfile} does not correspond "
                        "with number of grid layers. Subgrid information will be removed. "
                        "This is a bug in xtgeo==2.15.2 for grids exported from RMS using Xtgeo. "
                        "Export the grid with xtgeo>2.15.2 to remove this warning. "
                    ),
                    FutureWarning,
                )
                grid.subgrids = None
            gridparameter = load_grid_parameter(grid, get_path(gridparameter))
            fence = get_fencespec(coords)
            hmin, hmax, vmin, vmax, values = grid.get_randomline(
                fence, gridparameter, zincrement=0.5
            )

            surface = xtgeo.surface_from_file(get_path(surfacepath))
            s_arr = get_surface_fence(fence, surface)
            return make_heatmap(
                values,
                s_arr=s_arr,
                theme=self.plotly_theme,
                s_name=self.surfacenames[self.surfacefiles.index(surfacepath)],
                colorscale=colorscale,
                xmin=hmin,
                xmax=hmax,
                ymin=vmin,
                ymax=vmax,
                zmin=color_values[0],
                zmax=color_values[1],
                xaxis_title="Distance along polyline",
                yaxis_title=self.zunit,
            )

        @app.callback(
            [
                Output(self.ids("color-values"), "min"),
                Output(self.ids("color-values"), "max"),
                Output(self.ids("color-values"), "value"),
                Output(self.ids("color-values"), "step"),
            ],
            [Input(self.ids("color-range-btn"), "n_clicks")],
            [State(self.ids("gridparameter"), "value")],
        )
        def _update_color_slider(_clicks, gridparameter):
            grid = load_grid(get_path(self.gridfile))
            gridparameter = load_grid_parameter(grid, get_path(gridparameter))

            minv = float(f"{gridparameter.values.min():2f}")
            maxv = float(f"{gridparameter.values.max():2f}")
            value = [minv, maxv]
            step = calculate_slider_step(minv, maxv, steps=100)
            return minv, maxv, value, step

    def add_webvizstore(self):
        return [
            (get_path, [{"path": fn}])
            for fn in [self.gridfile] + self.gridparafiles + self.surfacefiles
        ]


# pylint: disable=too-many-arguments, too-many-locals
def make_heatmap(
    arr,
    s_arr,
    theme,
    s_color="black",
    s_name=None,
    height=800,
    zmin=None,
    zmax=None,
    xmin=None,
    xmax=None,
    ymin=None,
    ymax=None,
    colorscale=None,
    uirevision=None,
    showscale=True,
    reverse_y=True,
    text=None,
    title=None,
    yaxis_title=None,
    xaxis_title=None,
):

    x_inc = (xmax - xmin) / arr.shape[1]
    y_inc = (ymax - ymin) / arr.shape[0]
    colors = (
        [[i / (len(colorscale) - 1), color] for i, color in enumerate(colorscale)]
        if colorscale
        else "RdBu"
    )
    layout = {}
    layout.update(theme["layout"])
    layout.update(
        {
            "height": height,
            "title": title,
            "uirevision": uirevision,
            "margin": {"b": 50, "t": 50, "r": 0},
            "yaxis": {
                "title": yaxis_title,
                "autorange": "reversed" if reverse_y else None,
            },
            "xaxis": {"title": xaxis_title},
        }
    )
    return {
        "data": [
            {
                "type": "heatmap",
                "name": "seismic",
                "text": text,
                "z": arr.tolist(),
                "x0": xmin,
                "xmax": xmax,
                "dx": x_inc,
                "y0": ymin,
                "ymax": ymax,
                "dy": y_inc,
                "zsmooth": "best",
                "showscale": showscale,
                "colorscale": colors,
                "zmin": zmin,
                "zmax": zmax,
            },
            {
                "type": "line",
                "y": s_arr[:, 2],
                "x": s_arr[:, 3],
                "name": s_name,
                "marker": {"color": s_color},
            },
        ],
        "layout": layout,
    }


@webvizstore
def get_path(path) -> Path:
    return Path(path)


def get_fencespec(coords):
    """Create a XTGeo fence spec from polyline coordinates"""
    poly = xtgeo.Polygons()
    poly.dataframe = pd.DataFrame(
        [
            {
                "X_UTME": c[1],
                "Y_UTMN": c[0],
                "Z_TVDSS": 0,
                "POLY_ID": 1,
                "NAME": "polyline",
            }
            for c in coords
        ]
    )
    return poly.get_fence(asnumpy=True)



from typing import Callable, Dict, List, Optional, Union

import webviz_core_components as wcc
import webviz_subsurface_components as wsc
from dash import html

from ..models import (
    PropertyStatisticsModel,
    ProviderTimeSeriesDataModel,
    SimulationTimeSeriesModel,
)
from .selector_view import (
    ensemble_selector,
    filter_selector,
    source_selector,
    vector_selector,
)


def surface_select_view(get_uuid: Callable, tab: str) -> html.Div:
    return html.Div(
        id=get_uuid("surface-select"),
        style={"width": "75%"},
        children=wcc.Dropdown(
            label="Surface statistics",
            id={"id": get_uuid("surface-type"), "tab": tab},
            options=[
                {"label": "Mean", "value": "mean"},
                {"label": "Standard Deviation", "value": "stddev"},
                {"label": "Minimum", "value": "min"},
                {"label": "Maximum", "value": "max"},
                {"label": "P10", "value": "p10"},
                {"label": "P90", "value": "p90"},
            ],
            clearable=False,
            value="mean",
        ),
    )


def surface_view(get_uuid: Callable, tab: str) -> html.Div:
    return html.Div(
        style={"height": "35vh"},
        children=[
            wcc.Label(
                id={"id": get_uuid("surface-name"), "tab": tab},
                children="Select vector, then click on a correlation to visualize surface",
            ),
            wsc.LeafletMap(
                id={"id": get_uuid("surface-view"), "tab": tab},
                layers=[],
                unitScale={},
                autoScaleMap=True,
                minZoom=-19,
                updateMode="update",
                mouseCoords={"position": "bottomright"},
                colorBar={"position": "bottomleft"},
            ),
        ],
    )


def timeseries_view(get_uuid: Callable) -> html.Div:
    return html.Div(
        style={"height": "38vh"},
        children=[
            wcc.Graph(
                id=get_uuid("property-response-vector-graph"),
                config={"displayModeBar": False},
                style={"height": "35vh"},
            ),
        ],
    )


def correlation_view(get_uuid: Callable) -> wcc.Graph:
    return html.Div(
        [
            wcc.Graph(
                style={"height": "38vh"},
                id=get_uuid("property-response-correlation-graph"),
            ),
            wcc.Graph(
                style={"height": "38vh", "margin-top": "20px"},
                id=get_uuid("property-response-scatter-graph"),
            ),
        ]
    )


def filter_correlated_parameter(get_uuid: Callable, labels: List[str]) -> html.Div:
    return html.Div(
        [
            html.Div(
                style={"marginBottom": "10px"},
                children=wcc.Dropdown(
                    label="Filter on property",
                    id=get_uuid("property-response-correlated-filter"),
                    options=[{"label": label, "value": label} for label in labels],
                    placeholder="Select a label to filter on...",
                ),
            ),
            wcc.RangeSlider(
                id=get_uuid("property-response-correlated-slider"),
                disabled=True,
                marks=None,
            ),
        ]
    )


def selector_view(
    get_uuid: Callable,
    property_model: PropertyStatisticsModel,
    vector_model: Union[SimulationTimeSeriesModel, ProviderTimeSeriesDataModel],
    surface_folders: Optional[Dict],
) -> html.Div:
    return html.Div(
        children=[
            wcc.Selectors(
                label="Selectors",
                children=[
                    ensemble_selector(
                        get_uuid=get_uuid,
                        ensembles=property_model.ensembles,
                        tab="response",
                    ),
                    source_selector(
                        get_uuid=get_uuid,
                        sources=property_model.sources,
                        tab="response",
                    ),
                    vector_selector(get_uuid=get_uuid, vector_model=vector_model),
                ],
            ),
            wcc.Selectors(
                label="Filters",
                children=[
                    filter_selector(
                        get_uuid=get_uuid, property_model=property_model, tab="response"
                    ),
                    filter_correlated_parameter(
                        get_uuid=get_uuid, labels=property_model.get_labels()
                    ),
                ],
            ),
            wcc.Selectors(
                label="Surface",
                children=[surface_select_view(get_uuid=get_uuid, tab="response")]
                if surface_folders is not None
                else [html.Div()],
            ),
        ],
    )


def property_response_view(
    get_uuid: Callable,
    property_model: PropertyStatisticsModel,
    vector_model: Union[SimulationTimeSeriesModel, ProviderTimeSeriesDataModel],
    surface_folders: Optional[Dict],
) -> wcc.FlexBox:
    return wcc.FlexBox(
        style={"margin": "20px"},
        children=[
            wcc.FlexColumn(
                children=wcc.Frame(
                    style={"height": "80vh", "overflowY": "auto"},
                    children=selector_view(
                        get_uuid=get_uuid,
                        property_model=property_model,
                        vector_model=vector_model,
                        surface_folders=surface_folders,
                    ),
                )
            ),
            wcc.FlexColumn(
                flex=4,
                children=wcc.Frame(
                    style={"height": "80vh"},
                    color="white",
                    highlight=False,
                    children=[
                        html.Div(
                            style={"height": "38vh"},
                            children=timeseries_view(get_uuid=get_uuid),
                        ),
                        html.Div(
                            style={"height": "39vh"},
                            children=surface_view(get_uuid=get_uuid, tab="response")
                            if surface_folders is not None
                            else None,
                        ),
                    ],
                ),
            ),
            wcc.FlexColumn(
                flex=4,
                children=wcc.Frame(
                    style={"height": "80vh"},
                    color="white",
                    highlight=False,
                    children=correlation_view(get_uuid=get_uuid),
                ),
            ),
        ],
    )



import numpy as np
import pandas as pd
import pytest
from pandas._testing import assert_frame_equal

from webviz_subsurface._datainput.well_completions import remove_invalid_colors
from webviz_subsurface.plugins._well_completions._business_logic import (
    extract_stratigraphy,
    merge_compdat_and_connstatus,
)


def test_remove_invalid_colors():
    """Tests that colors that are not 6 digit hexadecimal are removed from the lyr
    parse zone list
    """
    zonelist = [
        {"name": "ZoneA", "color": "#FFFFFF"},
        {"name": "ZoneB", "color": "#FFF"},
    ]
    assert remove_invalid_colors(zonelist) == [
        {"name": "ZoneA", "color": "#FFFFFF"},
        {
            "name": "ZoneB",
        },
    ]


def test_extract_stratigraphy():
    """Checks that the merging of the layer_zone_mapping and the stratigraphy is
    correct and that the colors are added following the correct prioritization
    rules
    """
    zone_names = ["ZoneA.1", "ZoneA.2", "ZoneB.1"]
    stratigraphy = [
        {
            "name": "ZoneA",
            "color": "#000000",
            "subzones": [{"name": "ZoneA.1"}, {"name": "ZoneA.2"}, {"name": "ZoneA.3"}],
        },
        {
            "name": "ZoneB",
            "subzones": [{"name": "ZoneB.1", "subzones": [{"name": "ZoneB.1.1"}]}],
        },
        {
            "name": "ZoneC",
        },
    ]
    zone_color_mapping = {"ZoneA": "#111111", "ZoneA.1": "#222222"}
    theme_colors = ["#FFFFFF"]

    result = extract_stratigraphy(
        zone_names, stratigraphy, zone_color_mapping, theme_colors
    )
    assert result == [
        {
            "name": "ZoneA",
            "color": "#000000",  # colors given in the stratigraphy has priority
            "subzones": [
                {
                    "name": "ZoneA.1",
                    "color": "#222222",  # color from zone_color_mapping
                },
                {"name": "ZoneA.2", "color": "#FFFFFF"}  # color from theme_colors
                # ZoneA.3 is not here because it is not in the layer_zone_mapping
            ],
        },
        {
            "name": "ZoneB",
            "color": "#808080",  # Since it's not a leaf, color is set to grey
            "subzones": [
                {
                    "name": "ZoneB.1",
                    "color": "#FFFFFF"  # color from theme_colors
                    # No subzones here because ZoneB.1.1 is not in the layer_zone_mapping
                }
            ],
        },
        # ZoneC is removed since it's not in the layer_zone_mapping
    ]


def test_extract_stratigraphy_errors():
    """Checks that a ValueError is raised when a Zone is in the layer_zone_mapping, but
    not in the stratigraphy.
    """
    layer_zone_mapping = {1: "ZoneA", 2: "ZoneB", 3: "ZoneD"}
    stratigraphy = [
        {
            "name": "ZoneA",
        },
        {
            "name": "ZoneB",
        },
    ]
    with pytest.raises(ValueError):
        extract_stratigraphy(layer_zone_mapping, stratigraphy, {}, [])


def test_merge_compdat_and_connstatus():
    """Tests the functionality of the merge_compdat_and_connstatus function.

    The following functionality is covered:
    * The two first rows of df_compdat is replaced with the two first from
    df_connstatus, except for the KH (which is only available in df_compdat).
    KH is taken from the first of the two compdat rows
    * The A2 well is not available in df_connstatus and will be taken as is
    from df_compdat
    * the fourth row in df_compdat (WELL: A1, REAL:1) is ignored because A1 is
    in df_connstatus, but not REAL 1. We don't mix compdat and connstatus data
    for the same well
    * The fourth row in df_compdat has KH=Nan. This will be 0 in the output
    """
    df_compdat = pd.DataFrame(
        data={
            "DATE": [
                "2021-01-01",
                "2021-05-01",
                "2021-01-01",
                "2021-01-01",
                "2022-01-01",
            ],
            "REAL": [0, 0, 0, 1, 0],
            "WELL": ["A1", "A1", "A2", "A1", "A3"],
            "I": [1, 1, 1, 1, 1],
            "J": [1, 1, 1, 1, 1],
            "K1": [1, 1, 1, 1, 1],
            "OP/SH": ["SHUT", "OPEN", "OPEN", "OPEN", "OPEN"],
            "KH": [100, 1000, 10, 100, np.nan],
        }
    )
    df_connstatus = pd.DataFrame(
        data={
            "DATE": ["2021-03-01", "2021-08-01", "2021-01-01"],
            "REAL": [0, 0, 0],
            "WELL": ["A1", "A1", "A3"],
            "I": [1, 1, 1],
            "J": [1, 1, 1],
            "K1": [1, 1, 1],
            "OP/SH": ["OPEN", "SHUT", "OPEN"],
        }
    )
    df_output = pd.DataFrame(
        data={
            "DATE": ["2021-03-01", "2021-08-01", "2021-01-01", "2021-01-01"],
            "REAL": [0, 0, 0, 0],
            "WELL": ["A1", "A1", "A3", "A2"],
            "I": [1, 1, 1, 1],
            "J": [1, 1, 1, 1],
            "K1": [1, 1, 1, 1],
            "OP/SH": ["OPEN", "SHUT", "OPEN", "OPEN"],
            "KH": [100.0, 100.0, 0.0, 10.0],
        }
    )
    df_result = merge_compdat_and_connstatus(df_compdat, df_connstatus)
    assert_frame_equal(
        df_result, df_output, check_like=True
    )  # Ignore order of rows and columns



import pandas as pd

from webviz_subsurface._components.parameter_filter import ParameterFilter


def test_dataframe(testdata_folder) -> None:

    dframe = pd.read_csv(
        testdata_folder / "reek_test_data" / "aggregated_data" / "parameters.csv"
    )
    component = ParameterFilter("test", dframe)
    # pylint: disable=protected-access
    assert set(component._discrete_parameters) == set(
        ["FWL", "MULTFLT_F1", "INTERPOLATE_WO", "COHIBA_MODEL_MODE", "RMS_SEED"]
    )
    assert component.is_sensitivity_run is True



# pylint: disable=no-name-in-module
from webviz_config.plugins import TornadoPlotterFMU


def test_tornado_plotter_fmu(dash_duo, app, shared_settings) -> None:
    plugin = TornadoPlotterFMU(
        shared_settings["SENS_SETTINGS"],
        ensemble=shared_settings["SENS_ENSEMBLES"][0],
        csvfile="share/results/volumes/geogrid--vol.csv",
        multi_value_selectors=["REGION", "ZONE"],
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



import sys
import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
import yaml
from webviz_config.utils import terminal_colors


def set_simulation_line_shape_fallback(line_shape_fallback: str) -> str:
    """
    Defines a Plotly line_shape to use if a vector is not thought to be a rate or a total.
    """
    line_shape_fallback = line_shape_fallback.lower()
    if line_shape_fallback in ("backfilled", "backfill"):
        return "vh"
    if line_shape_fallback in ["hv", "vh", "hvh", "vhv", "spline", "linear"]:
        return line_shape_fallback
    warnings.warn(
        f"{line_shape_fallback}, is not a valid line_shape option, will use linear."
    )
    return "linear"


def get_simulation_line_shape(
    line_shape_fallback: str, vector: str, smry_meta: Optional[pd.DataFrame] = None
) -> str:
    """Get simulation time series line shape, smry_meta is a pd.DataFrame on the format given
    by `load_smry_meta` in `../_datainput/fmu_input.py`.
    """
    if vector.startswith("AVG_") or vector.startswith("INTVL_"):
        # These custom calculated vectors are valid forwards in time.
        return "hv"

    if smry_meta is None:
        return line_shape_fallback
    try:
        if smry_meta.is_rate[vector]:
            # Eclipse rate vectors are valid backwards in time.
            return "vh"
        if smry_meta.is_total[vector]:
            return "linear"
    except (AttributeError, KeyError):
        pass
    return line_shape_fallback


def calc_series_statistics(
    df: pd.DataFrame, vectors: list, refaxis: str = "DATE"
) -> pd.DataFrame:
    """Calculate statistics for given vectors over the ensembles
    refaxis is used if another column than DATE should be used to groupby.
    """
    # Invert p10 and p90 due to oil industry convention.
    def p10(x: List[float]) -> np.floating:
        return np.nanpercentile(x, q=90)

    def p90(x: List[float]) -> np.floating:
        return np.nanpercentile(x, q=10)

    def p50(x: List[float]) -> np.floating:
        return np.nanpercentile(x, q=50)

    # Calculate statistics, ignoring NaNs.
    stat_df = (
        df[["ENSEMBLE", refaxis] + vectors]
        .groupby(["ENSEMBLE", refaxis])
        .agg([np.nanmean, np.nanmin, np.nanmax, p10, p90, p50])
        .reset_index(level=["ENSEMBLE", refaxis], col_level=1)
    )
    # Rename nanmin, nanmax and nanmean to min, max and mean.
    col_stat_label_map = {
        "nanmin": "min",
        "nanmax": "max",
        "nanmean": "mean",
        "p10": "high_p10",
        "p90": "low_p90",
        "p50": "p50",
    }
    stat_df.rename(columns=col_stat_label_map, level=1, inplace=True)

    return stat_df


def add_statistics_traces(
    ens_stat_df: pd.DataFrame,
    vector: str,
    stat_options: List[str],
    color: str,
    legend_group: str,
    line_shape: str,
    refaxis: str = "DATE",
    hovertemplate: str = "(%{x}, %{y})<br>",
) -> List[Dict[str, Any]]:
    """Renders a statistical lines for each vector"""
    traces = []
    for i, opt in enumerate(stat_options):
        if opt == "Mean":
            traces.append(
                {
                    "name": legend_group,
                    "hovertemplate": hovertemplate + "Mean",
                    "x": ens_stat_df[("", refaxis)],
                    "y": ens_stat_df[(vector, "mean")],
                    "mode": "lines",
                    "line": {"color": color, "shape": line_shape, "width": 3},
                    "legendgroup": legend_group,
                    "showlegend": True,
                }
            )
        if opt == "P10 (high)":
            traces.append(
                {
                    "name": legend_group,
                    "hovertemplate": hovertemplate + "P10",
                    "x": ens_stat_df[("", refaxis)],
                    "y": ens_stat_df[(vector, "high_p10")],
                    "mode": "lines",
                    "line": {"color": color, "shape": line_shape, "dash": "dashdot"},
                    "legendgroup": legend_group,
                    "showlegend": False if "Mean" in stat_options else i == 0,
                }
            )
        if opt == "P90 (low)":
            traces.append(
                {
                    "name": legend_group,
                    "hovertemplate": hovertemplate + "P90",
                    "x": ens_stat_df[("", refaxis)],
                    "y": ens_stat_df[(vector, "low_p90")],
                    "mode": "lines",
                    "line": {"color": color, "shape": line_shape, "dash": "dashdot"},
                    "legendgroup": legend_group,
                    "showlegend": False if "Mean" in stat_options else i == 0,
                }
            )
        if opt == "P50 (median)":
            traces.append(
                {
                    "name": legend_group,
                    "hovertemplate": hovertemplate + "P50",
                    "x": ens_stat_df[("", refaxis)],
                    "y": ens_stat_df[(vector, "p50")],
                    "mode": "lines",
                    "line": {
                        "color": color,
                        "shape": line_shape,
                        "dash": "dot",
                        "width": 3,
                    },
                    "legendgroup": legend_group,
                    "showlegend": False if "Mean" in stat_options else i == 0,
                }
            )
        if opt == "Maximum":
            traces.append(
                {
                    "name": legend_group,
                    "hovertemplate": hovertemplate + "Maximum",
                    "x": ens_stat_df[("", refaxis)],
                    "y": ens_stat_df[(vector, "max")],
                    "mode": "lines",
                    "line": {
                        "color": color,
                        "shape": line_shape,
                        "dash": "longdash",
                        "width": 1.5,
                    },
                    "legendgroup": legend_group,
                    "showlegend": False if "Mean" in stat_options else i == 0,
                }
            )
        if opt == "Minimum":
            traces.append(
                {
                    "name": legend_group,
                    "hovertemplate": hovertemplate + "Minimum",
                    "x": ens_stat_df[("", refaxis)],
                    "y": ens_stat_df[(vector, "min")],
                    "mode": "lines",
                    "line": {
                        "color": color,
                        "shape": line_shape,
                        "dash": "longdash",
                        "width": 1.5,
                    },
                    "legendgroup": legend_group,
                    "showlegend": False if "Mean" in stat_options else i == 0,
                }
            )
    return traces


def render_hovertemplate(vector: str, interval: Optional[str]) -> str:
    if vector.startswith(("AVG_", "INTVL_")) and interval is not None:
        if interval == "daily":
            return "(%{x|%b} %{x|%-d}, %{x|%Y}, %{y})<br>"
        if interval == "monthly":
            return "(%{x|%b} %{x|%Y}, %{y})<br>"
        if interval == "yearly":
            return "(%{x|%Y}, %{y})<br>"
        raise ValueError(f"Interval {interval} is not supported.")
    return "(%{x}, %{y})<br>"  # Plotly's default behavior


def date_to_interval_conversion(
    date: Optional[str], vector: str, interval: str, as_date: bool = False
) -> Optional[str]:
    """Converts a date on form YYYY-MM-DD to an interval 'date' if the
    vector is a vector that is calculated from cumulatives (AVG_ or INTVL_).
    If as_date=True, the date returned will be the the first date of the interval
    (independent of which date in the interval the input is),
    if not, the 'date' returned will be the common basis for the interval,
    e.g. YYYY-MM-DD for daily, YYYY-MM for monthly and YYYY for yearly.

    The input date must be a date that can be reduced to the interval date by simply
    removing terms, e.g. is both 2001-05-16 and 2001-05-01 ok for monthly and will return
    2001-05.
    """
    if date is None:
        return date
    if vector.startswith(("AVG_", "INTVL_")):
        if interval == "monthly":
            date = "-".join(date.split("-")[0:2]) + ("-01" if as_date else "")
        if interval == "yearly":
            date = date.split("-")[0] + ("-01-01" if as_date else "")
    return date


# pylint: disable=too-many-branches
def check_and_format_observations(obsfile: Path) -> dict:
    with open(obsfile, "r") as stream:
        try:
            obsfile_data = yaml.safe_load(stream)
        except yaml.MarkedYAMLError as excep:
            extra_info = (
                f"There is something wrong in the configuration file {obsfile}. "
            )

            if (problem_mark := getattr(excep, "problem_mark", None)) is not None:
                extra_info += (
                    "The typo is probably somewhere around "
                    f"line {problem_mark.line + 1}."
                )

            raise type(excep)(
                f"{excep}. {terminal_colors.RED}{terminal_colors.BOLD}"
                f"{extra_info}{terminal_colors.END}"
            ).with_traceback(sys.exc_info()[2])

        try:
            obslist = obsfile_data.pop("smry")
        except KeyError as exc:
            raise KeyError(
                "The observation file lacks a `smry` section, which is mandatory for observations "
                "to work with this plugin."
            ) from exc
        except TypeError as exc:
            raise TypeError(
                "The observation file's othermost level must be a dictionary, while the input "
                f"file's outermost level is of type {type(obsfile_data)}."
            ) from exc
        if not isinstance(obslist, list):
            raise TypeError(
                "The observation file's smry section must be formatted as a list of dictionaries."
            )
        observations: dict = {}

        # pylint: disable=too-many-nested-blocks
        for item in obslist:
            if isinstance(item, dict):
                if "key" not in item:
                    raise KeyError(
                        "Missing mandatory element `key` in the observation file's smry section "
                        f"for the entry {item}."
                    )
                if "observations" not in item:
                    raise KeyError(
                        "Missing mandatory element `observations` in the observation file's smry "
                        f"section the entry {item}."
                    )
                key = item.pop("key")
                if isinstance(item["observations"], list):
                    for obs in item["observations"]:
                        if isinstance(obs, dict):
                            if "value" in obs and "date" in obs:
                                # Silently ignoring missing error, as the observation is still
                                # plotted as a point as long as "value" and "date"
                                # are present.
                                pass
                            else:
                                raise KeyError(
                                    f"Missing value and/or date for smry observation {obs} "
                                    f"under key {key} in smry section of observation file."
                                )
                        else:
                            raise TypeError(
                                f"Observation not formatted as dictionary in smry section of "
                                f"observation file. Check key: {key}, observation: {obs}."
                            )
                else:
                    raise KeyError(
                        "A smry observation must at least contain a `value` and a `date`."
                        f"Check at least key: {key}, observation: {item} in the observation "
                        f"file."
                    )
                observations[key] = item
            else:
                raise TypeError(
                    f"{item} in the observation file's smry section is not a dictionary."
                )
    return observations
